"""
3D layouts
==========

This example shows how to visualise graphs or networks in 3D using `iplotx`. Of course, a 3D layout is needed
for this. Here, we use the Fruchterman-Reingold layout algorithm from ``igraph`` to generate a 3D layout.

.. note::
    3D visualisation is most useful when used **interactively**, so you can rotate and pan the plot to inspect
    it from different angles. Matplotlib supports this both in Jupyter notebooks and in IPython via
    multiple interactive backends (e.g., TkAgg, Qt5Agg, etc.). These plots can also be saved as static
    images (the ones you see below were generated this way), however these static images can be quite
    difficult to interpret.
"""

import igraph as ig
import iplotx as ipx

# Make the graph
g = ig.Graph.Erdos_Renyi(30, m=50)

# Make a 3D layout
layout = g.layout_fruchterman_reingold_3d()

# Visualise the graph in 3D
ipx.network(
    g,
    layout,
    vertex_alpha=0.7,
    edge_alpha=0.4,
    figsize=(8, 8),
)

# %%
# Below is a variation using arrows and vertex labels:

import igraph as ig
import iplotx as ipx

# Make the graph
g = ig.Graph.Erdos_Renyi(30, m=50, directed=True)

# Make a 3D layout
layout = g.layout_fruchterman_reingold_3d()

# Visualise the graph in 3D
ipx.network(
    g,
    layout,
    vertex_alpha=0.3,
    edge_alpha=0.5,
    vertex_labels=True,
    figsize=(8, 8),
)

# %%
# .. warning::
#    3D visualisation does not support all features of 2D visualisation yet. Curved edges, waypoints, and edge
#    labels are currently unsupported. PRs are welcome!



"""
All artists defined in iplotx.
"""

from .network import NetworkArtist
from .tree import TreeArtist
from .vertex import VertexCollection
from .edge import EdgeCollection
from .label import LabelCollection
from .edge.arrow import EdgeArrowCollection
from .edge.leaf import LeafEdgeCollection
from .cascades import CascadeCollection
from .art3d.vertex import Vertex3DCollection
from .art3d.edge import Edge3DCollection


___all__ = (
    NetworkArtist,
    TreeArtist,
    VertexCollection,
    EdgeCollection,
    LeafEdgeCollection,
    LabelCollection,
    EdgeArrowCollection,
    CascadeCollection,
    Vertex3DCollection,
    Edge3DCollection,
)



"""
Leaf edges
==========

This example shows how to use leaf edges, dashed lines connecting each leaf to the
deepest leaf and label.
"""

from ete4 import Tree
import iplotx as ipx

tree = Tree(
    "((),((),(((),()),((),()))));",
)

ipx.plotting.tree(
    tree,
    leaf_labels={leaf: str(i + 1) for i, leaf in enumerate(tree.leaves())},
)

# %%
# Leaf edges are used by default when leaf labels are present and deep
# leaves are set (default). If you use shallow leaves, leaf edges are
# not plotted:

ipx.plotting.tree(
    tree,
    leaf_labels={leaf: str(i + 1) for i, leaf in enumerate(tree.leaves())},
    leaf_deep=False,
)

# %%
# Similarly, if there are no leaf labels, leaf edges are not plotted:

ipx.plotting.tree(
    tree,
)

# %%
# You can finely tune this behaviour. For instance, you can force leaf edges
# despite absence of leaf labels, and change their visual appearance:

ipx.plotting.tree(
    tree,
    leaf_deep=True,
    leafedge_color="tomato",
    leafedge_linewidth=2,
    leafedge_linestyle="-.",
)



"""
Pollinators
===========

This example demonstrates how to visualise the bipartite relationship between plants and their pollinators using ``iplotx``.

This particular example uses igraph to load and process the network data, but you can also use networkx, the internal
data structures of ``iplotx``, or any other library you prefer.
"""

import igraph as ig
import pandas as pd
import matplotlib.pyplot as plt
import iplotx as ipx


# Load the data with the adjacency matrix. 1 means that pollinator j visits plant i.
edge_data = pd.read_csv("https://dshizuka.github.io/networkanalysis/SampleData/Sample_bipartite.csv", index_col=0)

# Convert it into a DataFrame for igraph
edge_data = edge_data.T.stack()
edge_data = edge_data[edge_data > 0].index.to_frame()
edge_data.columns = ["from", "to"]
g = ig.Graph.DataFrame(edge_data, directed=True, use_vids=False)

# Set vertex types: plants (0) and pollinators (1)
g.vs["type"] = [0 if name.startswith("Plant") else 1 for name in g.vs["name"]]

# Compute bipartite layout
layout = g.layout_bipartite()


fig, ax = plt.subplots(figsize=(7, 4))
ipx.network(
    g,
    layout=layout,
    style="hollow",
    ax=ax,
    vertex_labels=g.vs["name"],
)
fig.tight_layout()


# %%
# We can add a bit of color and invert the y axis direction for clarity:

color_map = {"hummingbird": "steelblue", "bat": "grey", "butterfly": "tomato", "moth": "brown", "bee": "gold"}
vertex_facecolors = {}
for i, name in enumerate(g.vs["name"]):
    vertex_facecolors[i] = color_map.get(name, "lightgreen")

edge_colors = []
for pollinator in edge_data["from"].values:
    edge_colors.append(color_map.get(pollinator, "lightgreen"))

# sphinx_gallery_thumbnail_number = 2
fig, ax = plt.subplots(figsize=(7, 4))
ipx.network(
    g,
    layout=layout,
    style="hollow",
    ax=ax,
    vertex_labels=g.vs["name"],
    vertex_facecolor=vertex_facecolors,
    vertex_alpha=0.6,
    edge_color=edge_colors,
)
ax.invert_yaxis()
fig.tight_layout()



"""
Vertex markers
==============

This example shows various kinds of vertex markers that can be used in `iplotx`.
"""

import igraph as ig
import numpy as np
import iplotx as ipx

markers = ["o", "s", "^", "v", "<", ">", "d", "e", "p", "h", "8", "*"]
n = len(markers)
g = ig.Graph(
    edges=[(0, i + 1) for i in range(n)],
    directed=False,
)

# Create star layout
layout = {i + 1: (np.cos(2 * np.pi / n * i), np.sin(2 * np.pi / n * i)) for i in range(n)}
layout[0] = (0, 0)

# Colors
colors = [
    "k",
    "#8dd3c7",
    "#ffffb3",
    "#bebada",
    "#fb8072",
    "#80b1d3",
    "#fdb462",
    "#b3de69",
    "#fccde5",
    "#bf5b17",
    "#666666",
    "#000000",
    "#bbbbbb",
]
colors = {i: colors[i] for i in range(n + 1)}

ipx.network(
    g,
    layout=layout,
    vertex_marker=markers,
    vertex_size={i: 20 * bool(i) for i in range(n + 1)},
    vertex_facecolor=colors,
    vertex_edgecolor="black",
    aspect=1,
)



"""
Typing module for data/object ingestion. This module described the abstract data types that providers need to comply with to be compatible with iplotx.

Networkx and trees are treated separately for practical reasons: many tree analysis libraries rely heavily on recursive data structures, which do not
work as well on general networks.
"""

from typing import (
    NotRequired,
    TypedDict,
    Protocol,
    Optional,
    Sequence,
    Any,
    Iterable,
)
from collections.abc import Hashable
import numpy as np
import pandas as pd
from ..typing import (
    GraphType,
    LayoutType,
    TreeType,
)
from .heuristics import (
    normalise_tree_layout,
)


class NetworkData(TypedDict):
    """Network data structure for iplotx."""

    directed: bool
    vertex_df: pd.DataFrame
    edge_df: pd.DataFrame
    ndim: int
    network_library: NotRequired[str]


class NetworkDataProvider(Protocol):
    """Protocol for network data ingestion provider for iplotx."""

    def __init__(
        self,
        network: GraphType,
    ) -> None:
        """Initialise network data provider.

        Parameters:
            network: The network to ingest.
        """
        self.network = network

    def __call__(
        self,
        layout: Optional[LayoutType] = None,
        vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
        edge_labels: Optional[Sequence[str] | dict] = None,
    ) -> NetworkData:
        """Create network data object for iplotx from any provider."""
        raise NotImplementedError("Network data providers must implement this method.")

    @staticmethod
    def check_dependencies():
        """Check whether the dependencies for this provider are installed."""
        raise NotImplementedError("Network data providers must implement this method.")

    @staticmethod
    def graph_type():
        """Return the graph type from this provider to check for instances."""
        raise NotImplementedError("Network data providers must implement this method.")

    def is_directed(self):
        """Check whether the network is directed."""
        raise NotImplementedError("Network data providers must implement this method.")

    def number_of_vertices(self):
        """The number of vertices/nodes in the network."""
        raise NotImplementedError("Network data providers must implement this method.")


class TreeData(TypedDict):
    """Tree data structure for iplotx."""

    rooted: bool
    directed: bool | str
    root: Optional[Hashable]
    leaf_df: pd.DataFrame
    vertex_df: dict[Hashable, tuple[float, float]]
    edge_df: dict[Hashable, Sequence[tuple[float, float]]]
    layout_coordinate_system: str
    layout_name: str
    orientation: str
    ndim: int
    tree_library: NotRequired[str]


class TreeDataProvider(Protocol):
    """Protocol for tree data ingestion provider for iplotx."""

    def __init__(
        self,
        tree: TreeType,
    ) -> None:
        """Initialize the provider with the tree type.

        Parameters:
            tree: The tree type that this provider will handle.
        """
        self.tree = tree

    @staticmethod
    def check_dependencies():
        """Check whether the dependencies for this provider are installed."""
        raise NotImplementedError("Tree data providers must implement this method.")

    @staticmethod
    def tree_type():
        """Return the tree type from this provider to check for instances."""
        raise NotImplementedError("Tree data providers must implement this method.")

    def is_rooted(self) -> bool:
        """Get whether the tree is rooted.

        Returns:
            A boolean indicating whether the tree is rooted.

        Note: This is a default implemntation that can be overridden by the provider
        if they support unrooted trees (e.g. Biopython).
        """
        return True

    def get_root(self) -> Any:
        """Get the tree root in a provider-specific data structure.

        Returns:
            The root of the tree.

        Note: This is a default implemntation that can be overridden by the provider.
        """
        if hasattr(self.tree, "root"):
            root_attr = self.tree.root
            if callable(root_attr):
                return root_attr()
            else:
                return root_attr
        return self.tree.get_root()

    def get_leaves(self) -> Sequence[Any]:
        """Get the tree leaves/tips in a provider-specific data structure.

        Returns:
            The leaves or tips of the tree.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    def preorder(self) -> Iterable[Any]:
        """Preorder (DFS - parent first) iteration over the tree.

        Returns:
            An iterable of nodes in preorder traversal.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    def postorder(self) -> Iterable[Any]:
        """Postorder (DFS - child first) iteration over the tree.

        Returns:
            An iterable of nodes in preorder traversal.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    @staticmethod
    def get_children(
        node: Any,
    ) -> Sequence[Any]:
        """Get the children of a node.

        Parameters:
            node: The node to get the children from.
        Returns:
            A sequence of children nodes.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    @staticmethod
    def get_branch_length(
        node: Any,
    ) -> Optional[float]:
        """Get the length of the branch to this node.

        Parameters:
            node: The node to get the branch length from.
        Returns:
            The branch length to the node.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    def get_branch_length_default_to_one(
        self,
        node: Any,
    ) -> float:
        """Get the length of the branch to this node, defaulting to 1.0 if not available.

        Parameters:
            node: The node to get the branch length from.
        Returns:
            The branch length to the node, defaulting to 1.0 if not available.
        """
        branch_length = self.get_branch_length(node)
        return branch_length if branch_length is not None else 1.0

    def get_lca(
        self,
        nodes: Sequence[Hashable],
    ) -> Hashable:
        """Find the last common ancestor of a sequence of nodes.

        Parameters:
            nodes: The nodes to find a common ancestor for.

        Returns:
            The node that is the last (deepest) common ancestor of the nodes.

        NOTE: individual providers may implement more efficient versions of
        this function if desired.
        """
        provider = self.__class__

        # Find leaves of the selected nodes
        leaves = set()
        for node in nodes:
            # NOTE: get_leaves excludes the node itself...
            if len(self.get_children(node)) == 0:
                leaves.add(node)
            else:
                leaves |= set(provider(node).get_leaves())

        # Look for nodes with the same set of leaves, starting from the bottom
        # and stopping at the first (i.e. lowest) hit.
        for node in self.postorder():
            # NOTE: As above, get_leaves excludes the node itself
            if len(self.get_children(node)) == 0:
                leaves_node = {node}
            else:
                leaves_node = set(provider(node).get_leaves())
            if leaves <= leaves_node:
                root = node
                break
        else:
            raise ValueError(f"Common ancestor not found for nodes: {nodes}")

        return root

    def __call__(
        self,
        layout: str | LayoutType,
        layout_style: Optional[dict[str, int | float | str]] = None,
        directed: bool = False,
        vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series | bool] = None,
        edge_labels: Optional[Sequence[str] | dict] = None,
        leaf_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series | bool] = None,
    ) -> TreeData:
        """Create tree data object for iplotx from any tree provider.

        NOTE: This function needs NOT be implemented by individual providers.
        """

        if layout_style is None:
            layout_style = {}

        orientation = layout_style.pop("orientation", None)
        if orientation is None:
            if layout == "horizontal":
                orientation = "right"
            elif layout == "vertical":
                orientation = "descending"
            elif layout == "radial":
                orientation = "clockwise"

        tree_data = {
            "root": self.get_root(),
            "rooted": self.is_rooted(),
            "directed": directed,
            "ndim": 2,
            "layout_name": layout,
            "orientation": orientation,
        }

        # Add vertex_df including layout
        tree_data["vertex_df"] = normalise_tree_layout(
            layout,
            orientation=orientation,
            root=tree_data["root"],
            preorder_fun=self.preorder,
            postorder_fun=self.postorder,
            children_fun=self.get_children,
            branch_length_fun=self.get_branch_length_default_to_one,
            **layout_style,
        )
        if layout in ("radial",):
            tree_data["layout_coordinate_system"] = "polar"
        else:
            tree_data["layout_coordinate_system"] = "cartesian"

        # Add leaf_df
        # NOTE: Sometimes (e.g. cogent3) the leaves convert into a pd.Index
        # in a strange way, whereby their name disappears upon printing the
        # index but is actually visible (and kept) when inspecting the
        # individual elements (leaves). Seems ok functionally, though a little
        # awkward visually during debugging.
        tree_data["leaf_df"] = pd.DataFrame(index=self.get_leaves())
        leaf_name_attrs = ("name",)

        # Add edge_df
        edge_data = {"_ipx_source": [], "_ipx_target": [], "branch_length": []}
        for node in self.preorder():
            for child in self.get_children(node):
                edge_data["_ipx_source"].append(node)
                edge_data["_ipx_target"].append(child)
                edge_data["branch_length"].append(self.get_branch_length(child))
        edge_df = pd.DataFrame(edge_data)
        tree_data["edge_df"] = edge_df

        # Add edge labels
        # NOTE: Partial support only for now, only lists
        if edge_labels is not None:
            # Cycling sequence
            edge_df["label"] = [edge_labels[i % len(edge_labels)] for i in range(len(edge_df))]

        # Add branch support
        if hasattr(self, "get_support"):
            support = self.get_support()

            for key, value in support.items():
                # Leaves never show support, it's not a branching point
                if key in tree_data["leaf_df"].index:
                    support[key] = ""
                elif value is None:
                    support[key] = ""
                elif np.isscalar(value):
                    # Assume support is in percentage and round it to nearest integer.
                    support[key] = str(int(np.round(value, 0)))
                else:
                    # Apparently multiple supports are accepted in some XML format
                    support[key] = "/".join(str(int(np.round(v, 0))) for v in value)

            tree_data["vertex_df"]["support"] = pd.Series(support).loc[tree_data["vertex_df"].index]

        # Add vertex labels
        if vertex_labels is None:
            vertex_labels = False
        if np.isscalar(vertex_labels) and vertex_labels:
            tree_data["vertex_df"]["label"] = [x.name for x in tree_data["vertex_df"].index]
        elif not np.isscalar(vertex_labels):
            # If a dict-like object is passed, it can be incomplete (e.g. only the leaves):
            # we fill the rest with empty strings which are not going to show up in the plot.
            if isinstance(vertex_labels, pd.Series):
                vertex_labels = dict(vertex_labels)
            if isinstance(vertex_labels, dict):
                for vertex in tree_data["vertex_df"].index:
                    if vertex not in vertex_labels:
                        vertex_labels[vertex] = ""
            tree_data["vertex_df"]["label"] = pd.Series(vertex_labels)

        # Add leaf labels
        if leaf_labels is None:
            leaf_labels = False
        if np.isscalar(leaf_labels) and leaf_labels:
            leaf_labels = []
            for leaf in tree_data["leaf_df"].index:
                for name_attr in leaf_name_attrs:
                    if hasattr(leaf, name_attr):
                        label = getattr(leaf, name_attr)
                        break
                else:
                    raise ValueError(
                        "Could not find leaf name attribute.",
                    )
                leaf_labels.append(label)
            tree_data["leaf_df"]["label"] = leaf_labels
        elif not np.isscalar(leaf_labels):
            # Leaves are already in the dataframe in a certain order, so sequences are allowed
            if isinstance(leaf_labels, (list, tuple, np.ndarray)):
                leaf_labels = {
                    leaf: label for leaf, label in zip(tree_data["leaf_df"].index, leaf_labels)
                }
            # If a dict-like object is passed, it can be incomplete (e.g. only the leaves):
            # we fill the rest with empty strings which are not going to show up in the plot.
            if isinstance(leaf_labels, pd.Series):
                leaf_labels = dict(leaf_labels)
            if isinstance(leaf_labels, dict):
                for leaf in tree_data["leaf_df"].index:
                    if leaf not in leaf_labels:
                        leaf_labels[leaf] = ""
            tree_data["leaf_df"]["label"] = pd.Series(leaf_labels)

        return tree_data



"""
Module containing code to manipulate vertex visualisations in 3D, especially the Vertex3DCollection class.
"""

from typing import (
    Sequence,
)
import numpy as np
from matplotlib import (
    cbook,
)
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import (
    Path3DCollection,
    text_2d_to_3d,
)

from ..utils.matplotlib import (
    _forwarder,
)
from ..vertex import (
    VertexCollection,
)


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class Vertex3DCollection(VertexCollection, Path3DCollection):
    """Collection of vertex patches for plotting."""

    def _update_before_draw(self) -> None:
        """Update the collection before drawing."""
        # Set the sizes according to the current figure dpi
        VertexCollection._update_before_draw(self)

        if isinstance(self.axes, Axes3D) and hasattr(self, "do_3d_projection"):
            self.do_3d_projection()

    def draw(self, renderer) -> None:
        """Draw the collection of vertices in 3D.

        Parameters:
            renderer: The renderer to use for drawing.
        """
        with self._use_zordered_offset():
            with cbook._setattr_cm(self, _in_draw=True):
                VertexCollection.draw(self, renderer)


def vertex_collection_2d_to_3d(
    col: VertexCollection,
    zs: np.ndarray | float | Sequence[float] = 0,
    zdir: str = "z",
    depthshade: bool = True,
    axlim_clip: bool = False,
):
    """Convert a 2D VertexCollection to a 3D Vertex3DCollection.

    Parameters:
        col: The 2D VertexCollection to convert.
        zs: The z coordinate(s) to use for the 3D vertices.
        zdir: The axis to use as the z axis (default is "z").
        depthshade: Whether to aply depth shading (default is True).
        axlim_clip: Whether to clip the vertices to the axes limits (default is False).
    """
    if not isinstance(col, VertexCollection):
        raise TypeError("vertices must be a VertexCollection")

    col.__class__ = Vertex3DCollection
    col._offset_zordered = None
    col._depthshade = depthshade
    col._in_draw = False
    col.set_3d_properties(zs, zdir, axlim_clip)

    # Labels if present
    if col.get_labels() is not None:
        for z, art in zip(zs, col.get_labels()._labelartists):
            # zdir=None means the text is always horizontal facing the camera
            text_2d_to_3d(art, z, zdir=None, axlim_clip=axlim_clip)



import unittest
import pytest
import numpy as np
import pandas as pd
from matplotlib import patches
import matplotlib.pyplot as plt
import iplotx as ipx


def vertex_collection(**kwargs):
    if "layout" not in kwargs:
        kwargs["layout"] = pd.DataFrame(
            data=np.array([[0, 0], [1, 1]], np.float64),
            index=["A", "B"],
        )
    return ipx.artists.VertexCollection(**kwargs)


class VertexCollectionTest(unittest.TestCase):
    def setUp(self):
        self.vertex_collection = vertex_collection()

    def test_vertex_count(self):
        self.assertEqual(len(self.vertex_collection), 2)

    def test_vertex_names(self):
        pd.testing.assert_index_equal(self.vertex_collection.get_index(), pd.Index(["A", "B"]))

    def test_layout(self):
        expected_layout = pd.DataFrame(
            data=np.array([[0, 0], [1, 1]], np.float64), index=["A", "B"]
        )
        pd.testing.assert_frame_equal(self.vertex_collection._layout, expected_layout)

    def test_vertex_id(self):
        self.assertEqual(self.vertex_collection.get_vertex_id(0), "A")

    def test_get_style(self):
        with ipx.style.context({"vertex": {"size": 5}}):
            style = ipx.style.get_style(".vertex")
            vc = vertex_collection(style=style)
            assert vc.get_style() == style

    def test_get_sizes(self):
        with ipx.style.context({"vertex": {"size": 5}}):
            vc = vertex_collection(style=ipx.style.get_style(".vertex"))
            sizes = vc.get_sizes()
            np.testing.assert_array_equal(sizes, np.array([5, 5]))
            sizes_dpi = vc.get_sizes_dpi()
            np.testing.assert_array_equal(sizes_dpi, np.array([5, 5]))

    def test_set_sizes(self):
        self.vertex_collection.set_sizes(None)
        np.testing.assert_array_equal(
            self.vertex_collection.get_sizes(), np.array([], dtype=np.float64)
        )

    def test_get_layout(self):
        expected_layout = pd.DataFrame(
            data=np.array([[0, 0], [1, 1]], np.float64), index=["A", "B"]
        )
        pd.testing.assert_frame_equal(self.vertex_collection.get_layout(), expected_layout)

    def test_get_offset(self):
        np.testing.assert_array_equal(
            self.vertex_collection.get_offsets(), np.array([[0, 0], [1, 1]], np.float64)
        )
        np.testing.assert_array_equal(
            self.vertex_collection.get_offsets(ignore_layout=False),
            np.array([[0, 0], [1, 1]], np.float64),
        )

    def test_set_offsets(self):
        self.vertex_collection.set_offsets(np.array([[2, 2], [3, 3]], np.float64))
        np.testing.assert_array_equal(
            self.vertex_collection.get_offsets(),
            np.array([[2, 2], [3, 3]], np.float64),
        )

    def test_update_layout(self):
        vc = vertex_collection()
        vc._offsets = None
        vc._update_offsets_from_layout()
        np.testing.assert_array_equal(
            self.vertex_collection.get_offsets(), np.array([[0, 0], [1, 1]], np.float64)
        )

    def test_updat_layout_unsupported(self):
        vc = vertex_collection()
        vc._layout_coordinate_system = "unsupported"
        with pytest.raises(ValueError):
            vc._update_offsets_from_layout()

    def test_update_layout_polar(self):
        vc = vertex_collection()
        vc._layout_coordinate_system = "polar"
        vc._layout.iloc[:, 0] = [1, 1]
        vc._layout.iloc[:, 1] = [0, np.pi]
        vc._update_offsets_from_layout()
        np.testing.assert_array_almost_equal(
            vc.get_offsets(),
            np.array([[1, 0], [-1, 0]], np.float64),
        )

        vc._offsets = None
        vc._update_offsets_from_layout()
        np.testing.assert_array_almost_equal(
            vc.get_offsets(),
            np.array([[1, 0], [-1, 0]], np.float64),
        )

    @pytest.mark.filterwarnings("error")
    def test_nolabels(self):
        with ipx.style.context({"vertex": {"size": "label"}}):
            style = ipx.style.get_style(".vertex")
            with pytest.raises(UserWarning):
                vertex_collection(style=style)

    @pytest.mark.filterwarnings("ignore")
    def test_nolabels_ignorewarning(self):
        with ipx.style.context({"vertex": {"size": "label"}}):
            style = ipx.style.get_style(".vertex")
            vc = vertex_collection(style=style)
            sizes = vc.get_sizes()
            np.testing.assert_array_almost_equal(sizes, np.array([20] * 2, np.float64), decimal=2)

    def test_no_labels_check(self):
        assert self.vertex_collection._labels is None
        self.assertFalse(hasattr(self.vertex_collection, "_label_collection"))
        assert self.vertex_collection.get_labels() is None

    def test_stale(self):
        self.vertex_collection.stale = True
        self.assertTrue(self.vertex_collection.stale)
        self.vertex_collection.stale = False
        self.assertFalse(self.vertex_collection.stale)

    def test_update_labels(self):
        assert self.vertex_collection._update_labels() is None

    def test_draw_invisible(self):
        self.vertex_collection.set_visible(False)
        fig = plt.figure()
        try:
            assert self.vertex_collection.draw(fig.canvas.get_renderer()) is None
        finally:
            plt.close(fig)

    def test_draw_zerolength(self):
        vc = vertex_collection(layout=pd.DataFrame(data=np.array([[], []]).T, index=[]))
        fig = plt.figure()
        try:
            assert vc.draw(fig.canvas.get_renderer()) is None
        finally:
            plt.close(fig)


class VertexCollectionLabelsTest(unittest.TestCase):
    def setUp(self):
        with ipx.style.context({"vertex": {"size": "label"}}):
            self.vertex_collection = vertex_collection(
                labels=["hello", "world"],
                style=ipx.style.get_style(".vertex"),
            )

    def test_label_creation(self):
        labels = self.vertex_collection.get_labels()
        assert labels is not None
        assert isinstance(labels, ipx.artists.LabelCollection)

    def test_children(self):
        children = self.vertex_collection.get_children()
        assert isinstance(children, tuple)
        assert len(children) == 1
        assert isinstance(children[0], ipx.artists.LabelCollection)

    def test_set_figure(self):
        fig = plt.figure()
        self.vertex_collection.set_figure(fig)
        try:
            assert self.vertex_collection.get_figure() == fig
            assert self.vertex_collection.get_labels().get_figure() == fig
        finally:
            plt.close(fig)

    def test_update_labels(self):
        self.vertex_collection._layout_coordinate_system = "polar"
        assert self.vertex_collection._update_labels() is None


class VertexCollectionCmapTest(unittest.TestCase):
    def setUp(self):
        with ipx.style.context(
            {
                "vertex": {
                    "cmap": plt.cm.viridis,
                    "facecolor": [2, 3],
                },
            }
        ):
            style = ipx.style.get_style(".vertex")
            vc = vertex_collection(style=style)
        self.vertex_collection = vc

    def test_init_cmap(self):
        assert self.vertex_collection.get_cmap() == plt.cm.viridis


def test_shapes():
    for marker in ["o", "c", "s", "^", "v", "<", ">", "d", "p", "h", "8", "e", "*"]:
        patch, size = ipx.vertex.make_patch(marker=marker)
        assert size == 20
        assert isinstance(patch, patches.Patch)

    with pytest.raises(KeyError):
        ipx.vertex.make_patch(marker="invalid_marker")



"""
Simple path
===========

A minimal example of using `iplotx` together with `networkx`.
"""

import networkx as nx
import iplotx as ipx

G = nx.path_graph(8)
pos = nx.spring_layout(G, seed=47)  # Seed layout for reproducibility
ipx.network(G, layout=pos)

# %%
# We can change the color of the vertices and edges with a touch of style:

ipx.network(
    G,
    layout=pos,
    style={
        "vertex": {
            "facecolor": "lightblue",
            "edgecolor": "navy",
            "linewidth": 2.5,
        },
        "edge": {
            "color": "navy",
            "linewidth": 2.5,
        },
    },
)



"""
Four grids
==========

This example from networkx shows different stylings for the same grid.
"""

import matplotlib.pyplot as plt
import networkx as nx
import iplotx as ipx

G = nx.grid_2d_graph(4, 4)  # 4x4 grid

pos = nx.spring_layout(G, iterations=100, seed=39775)

# Create a 2x2 subplot
fig, all_axes = plt.subplots(2, 2)
axs = all_axes.flat

ipx.plot(
    G,
    layout=pos,
    ax=axs[0],
    style={
        "vertex": {
            "size": 25,
            "facecolor": "steelblue",
            "edgecolor": "none",
        }
    },
)

ipx.plot(
    G,
    layout=pos,
    ax=axs[1],
    style={
        "vertex": {
            "size": 0,
        }
    },
)

ipx.plot(
    G,
    layout=pos,
    ax=axs[2],
    style={
        "vertex": {
            "size": 25,
            "facecolor": "tab:green",
            "edgecolor": "tab:gray",
        },
        "edge": {
            "linewidth": 6,
            "color": "tab:gray",
        },
    },
)

H = G.to_directed()

ipx.plot(
    H,
    layout=pos,
    ax=axs[3],
    style={
        "vertex": {
            "facecolor": "tab:orange",
            "edgecolor": "none",
            "size": 10,
        },
        "edge": {
            "linewidth": 1.7,
            "offset": 0,
            "arrow": {
                "height": 5,
                "marker": "|>",
            },
        },
    },
)

# Set margins for the axes so that nodes aren't clipped
for ax in axs:
    ax.margins(0.10)
fig.tight_layout()



"""
Cogent3 tree
==============

This example shows how to use `iplotx` to plot trees from `cogent3`, the successor to `pycogent`.
"""

from cogent3.phylo import nj
import numpy as np
import iplotx as ipx
import matplotlib.pyplot as plt

nleaves = 8
distance_dict = {}
for i in range(nleaves):
    for j in range(i):
        distance_dict[(str(i), str(j))] = np.random.rand()
tree = nj.nj(distance_dict)
ipx.plotting.tree(
    tree,
)

# %%
# To add labels to the leaves, you can use the `vertex_labels` argument as a dictionary.
# Moreover, the plot can be customised further using hmargin (horizontal label margin)
# and some `matplotlib` settings.

leaf_labels = {leaf: f"Species {i + 1}" for i, leaf in enumerate(tree.tips())}
fig, ax = plt.subplots(figsize=(5, 4))
ipx.plotting.tree(
    tree,
    ax=ax,
    vertex_labels=leaf_labels,
    vertex_label_hmargin=10,
)
ax.set_xlim(-0.2, 1.5)

# %%
# `iplotx` can compute a radial tree layout as well, and usual style modifications
# work for trees same as networks:

ipx.plotting.tree(
    tree,
    layout="radial",
    style=[
        "tree",
        {
            "edge": {
                "color": "deeppink",
                "linewidth": 4,
            },
        },
    ],
    aspect=1,
)



from typing import (
    Any,
    Optional,
    Sequence,
)
import importlib
from functools import partialmethod

from ...typing import (
    TreeDataProvider,
)


class BiopythonDataProvider(TreeDataProvider):
    def is_rooted(self) -> bool:
        return self.tree.rooted

    def _traverse(self, order: str) -> Any:
        """Traverse the tree."""
        return self.tree.find_clades(order=order)

    preorder = partialmethod(_traverse, order="preorder")
    postorder = partialmethod(_traverse, order="postorder")

    def get_leaves(self) -> Sequence[Any]:
        return self.tree.get_terminals()

    @staticmethod
    def get_children(node: Any) -> Sequence[Any]:
        return node.clades

    @staticmethod
    def get_branch_length(node: Any) -> Optional[float]:
        return node.branch_length

    @staticmethod
    def check_dependencies() -> bool:
        return importlib.util.find_spec("Bio") is not None

    @staticmethod
    def tree_type():
        from Bio import Phylo

        return Phylo.BaseTree.Tree

    def get_support(self):
        """Get support/confidence values for all nodes."""
        support_dict = {}
        for node in self.preorder():
            if hasattr(node, "confidences"):
                support = node.confidences
            elif hasattr(node, "confidence"):
                support = node.confidence
            else:
                support = None
            support_dict[node] = support
        return support_dict



"""
Protein Protein Interactions
============================

This example demonstrates how to visualise Protein Protein Interaction (PPI) networks using ``iplotx``.

We use ``igraph`` to download the physical PPI network of *E. coli* from STRING database. The same thing could be done with ``networkx``.
We also subsample the network to just 300 random interactions to keep it reasonably quick, but different filtering strategies are possible.
"""

import igraph as ig
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import iplotx as ipx


# Load the data with the adjacency matrix. 1 means that pollinator j visits plant i.
edge_data = pd.read_csv("https://stringdb-downloads.org/download/protein.physical.links.v12.0/511145.protein.physical.links.v12.0.txt.gz", sep=" ")
edge_data["weight"] = 1.0 * edge_data["combined_score"] / edge_data["combined_score"].max()

# Subset just to keep it quick as an example
edge_data = edge_data.sample(300, random_state=42)

# Convert it into a DataFrame for igraph
g = ig.Graph.DataFrame(edge_data, directed=False, use_vids=False)

# Compute layout
layout = g.layout_fruchterman_reingold()

# Highlight hubs
degrees = np.array(g.vs.degree()).astype(float)
vertex_size = 3 + 27 * degrees / degrees.max()
vertex_alpha = np.clip(0.2 + 0.8 * degrees / degrees.max(), 0, 1)
vertex_color = ["red" if d > degrees.max() / 2 else "black" for d in degrees]

# Visualise
fig, ax = plt.subplots(figsize=(7, 7))
ipx.network(
    g,
    layout=layout,
    ax=ax,
    vertex_size=vertex_size,
    vertex_alpha=vertex_alpha,
    vertex_facecolor=vertex_color,
)
fig.tight_layout()



from typing import (
    Optional,
    Sequence,
    Self,
)
import numpy as np
import pandas as pd
import matplotlib as mpl

from .typing import (
    GraphType,
    LayoutType,
)
from .style import (
    get_style,
    rotate_style,
)
from .utils.matplotlib import (
    _stale_wrapper,
    _forwarder,
    _build_cmap_fun,
)
from .ingest import (
    ingest_network_data,
)
from .vertex import (
    VertexCollection,
)
from .edge import (
    EdgeCollection,
    make_stub_patch as make_undirected_edge_patch,
)
from .art3d.vertex import (
    vertex_collection_2d_to_3d,
)
from .art3d.edge import (
    Edge3DCollection,
    edge_collection_2d_to_3d,
)


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class NetworkArtist(mpl.artist.Artist):
    def __init__(
        self,
        network: GraphType,
        layout: Optional[LayoutType] = None,
        vertex_labels: Optional[list | dict | pd.Series] = None,
        edge_labels: Optional[Sequence] = None,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        offset_transform: Optional[mpl.transforms.Transform] = None,
    ):
        """Network container artist that groups all plotting elements.

        Parameters:
            network: The network to plot.
            layout: The layout of the network. If None, this function will attempt to
                infer the layout from the network metadata, using heuristics. If that fails, an
                exception will be raised.
            vertex_labels: The labels for the vertices. If None, no vertex labels
                will be drawn. If a list, the labels are taken from the list. If a dict, the keys
                should be the vertex IDs and the values should be the labels.
            edge_labels: The labels for the edges. If None, no edge labels will be drawn.
            transform: The transform to use for the vertices. Default is IdentityTransform.
            offset_transform: The transform to use as offset transform for the vertices and main
                transform for the edges. Default is None, but this should eventually be set to
                ax.transData once the artist is added to an Axes.

        """
        self.network = network

        super().__init__()

        # This is usually the identity (which scales poorly with dpi)
        self.set_transform(transform)

        # This is usually transData
        self.set_offset_transform(offset_transform)

        zorder = get_style(".network").get("zorder", 1)
        self.set_zorder(zorder)

        if network is not None:
            self._ipx_internal_data = ingest_network_data(
                network,
                layout,
                vertex_labels=vertex_labels,
                edge_labels=edge_labels,
            )

            self._add_vertices()
            self._add_edges()

    @classmethod
    def from_other(
        cls: "NetworkArtist",  # NOTE: This is fixed in Python 3.14
        other: Self,
    ) -> Self:
        """Create a NetworkArtist as a copy of another one.

        Parameters:
            other: The other NetworkArtist.

        Returns:
            An instantiated NetworkArtist.
        """
        self = cls.from_edgecollection(other._edges)
        self.network = other.network
        if hasattr(other, "_ipx_internal_data"):
            self._ipx_internal_data = other._ipx_internal_data
        return self

    @classmethod
    def from_edgecollection(
        cls: "NetworkArtist",  # NOTE: This is fixed in Python 3.14
        edge_collection: EdgeCollection | Edge3DCollection,
    ) -> Self:
        """Create a NetworkArtist from iplotx artists.

        Parameters:
            edge_collection: The edge collection to use to initialise the artist. Vertices will
              be obtained automatically.

        Returns:
            The initialised NetworkArtist.
        """
        vertex_collection = edge_collection._vertex_collection
        layout = vertex_collection._layout
        transform = vertex_collection.get_transform()
        offset_transform = vertex_collection.get_offset_transform()

        # Follow the steps in the normal constructor
        self = cls(
            network=None,
            layout=layout,
            transform=transform,
            offset_transform=offset_transform,
        )
        # TODO: should we make copies here?
        self._vertices = vertex_collection
        self._edges = edge_collection

        return self

    def get_children(self):
        if hasattr(self, "_edges"):
            return (self._vertices, self._edges)
        else:
            return (self._vertices,)

    def set_figure(self, fig):
        super().set_figure(fig)
        for child in self.get_children():
            child.set_figure(fig)

    @property
    def axes(self):
        return mpl.artist.Artist.axes.__get__(self)

    @axes.setter
    def axes(self, new_axes):
        mpl.artist.Artist.axes.__set__(self, new_axes)
        for child in self.get_children():
            child.axes = new_axes
        self.set_figure(new_axes.figure)

    def get_offset_transform(self):
        """Get the offset transform (for vertices/edges)."""
        return self._offset_transform

    def set_offset_transform(self, offset_transform):
        """Set the offset transform (for vertices/edges)."""
        self._offset_transform = offset_transform
        if hasattr(self, "_vertices"):
            self._vertices.set_offset_transform(offset_transform)
        if hasattr(self, "_edges"):
            self._edges.set_transform(offset_transform)

    def get_vertices(self):
        """Get VertexCollection artist."""
        return self._vertices

    get_nodes = get_vertices

    def get_edges(self):
        """Get EdgeCollection artist."""
        return self._edges

    def get_vertex_labels(self):
        """Get list of vertex label artists."""
        return self._vertices.get_labels()

    get_node_labels = get_vertex_labels

    def get_edge_labels(self):
        """Get list of edge label artists."""
        return self._edges.get_labels()

    def get_datalim(self, transData, pad=0.15):
        """Get limits on x/y axes based on the graph layout data.

        Parameters:
            transData (Transform): The transform to use for the data.
            pad (float): Padding to add to the limits. Default is 0.05.
                Units are a fraction of total axis range before padding.
        """
        layout = self.get_layout().values

        if len(layout) == 0:
            return mpl.transforms.Bbox([[0, 0], [1, 1]])

        bboxes = [
            self._vertices.get_datalim(transData),
        ]
        if hasattr(self, "_edges"):
            bboxes.append(
                self._edges.get_datalim(transData),
            )
        bbox = mpl.transforms.Bbox.union(bboxes)

        bbox = bbox.expanded(sw=(1.0 + pad), sh=(1.0 + pad))
        return bbox

    def autoscale_view(self, tight=False):
        """Recompute data limits from this artist and set autoscale based on them."""
        bbox = self.get_datalim(self.axes.transData)
        self.axes.update_datalim(bbox)
        self.axes.autoscale_view(tight=tight)

    def get_layout(self):
        """Get the vertex layout.

        Returns:
            The vertex layout as a DataFrame.
        """
        layout_columns = [f"_ipx_layout_{i}" for i in range(self.get_ndim())]
        vertex_layout_df = self._ipx_internal_data["vertex_df"][layout_columns]
        return vertex_layout_df

    def get_ndim(self):
        """Get the dimensionality of the layout.

        Returns:
            The dimensionality of the layout (2 or 3).
        """
        return self._ipx_internal_data["ndim"]

    def _get_label_series(self, kind):
        # Equivalence vertex/node
        if kind == "node":
            kind = "vertex"

        if "label" in self._ipx_internal_data[f"{kind}_df"].columns:
            return self._ipx_internal_data[f"{kind}_df"]["label"]
        else:
            return None

    def _add_vertices(self):
        """Add vertices to the network artist."""

        self._vertices = VertexCollection(
            layout=self.get_layout(),
            layout_coordinate_system=self._ipx_internal_data.get(
                "layout_coordinate_system", "cartesian"
            ),
            style=get_style(".vertex"),
            labels=self._get_label_series("vertex"),
            transform=self.get_transform(),
            offset_transform=self.get_offset_transform(),
        )

        if self.get_ndim() == 3:
            vertex_collection_2d_to_3d(
                self._vertices,
                zs=self.get_layout().iloc[:, 2].values,
                depthshade=False,
            )

    def _add_edges(self):
        """Add edges to the network artist.

        NOTE: UndirectedEdgeCollection and ArrowCollection are both subclasses of
        PatchCollection. When used with a cmap/norm, they set their facecolor
        according to the cmap, even though most likely we only want the edgecolor
        set that way. It can make for funny looking plots that are not uninteresting
        but mostly niche at this stage. Therefore we sidestep the whole cmap thing
        here.
        """

        labels = self._get_label_series("edge")
        edge_style = get_style(".edge")

        edge_df = self._ipx_internal_data["edge_df"].set_index(["_ipx_source", "_ipx_target"])

        if len(edge_df) == 0:
            return

        if "cmap" in edge_style:
            cmap_fun = _build_cmap_fun(
                edge_style,
                "color",
                edge_style.get("norm", None),
                internal=edge_df,
            )
        else:
            cmap_fun = None

        if "cmap" in edge_style:
            colorarray = []
        edgepatches = []
        adjacent_vertex_ids = []
        for i, (vid1, vid2) in enumerate(edge_df.index):
            edge_stylei = rotate_style(edge_style, index=i, key=(vid1, vid2))

            # FIXME:: Improve this logic. We have three layers of priority:
            # 1. Explicitely set in the style of "plot"
            # 2. Internal through network attributes
            # 3. Default styles
            # Because 1 and 3 are merged as a style context on the way in,
            # it's hard to squeeze 2 in the middle. For now, we will assume
            # the priority order is 2-1-3 instead (internal property is
            # highest priority).
            # This is also why we cannot shift this logic further into the
            # EdgeCollection class, which is oblivious of NetworkArtist's
            # internal data. In fact, one would argue this needs to be
            # pushed outwards to deal with the wrong ordering.
            _update_from_internal(edge_stylei, edge_df.iloc[i], kind="edge")

            if cmap_fun is not None:
                colorarray.append(edge_stylei["color"])
                edge_stylei["color"] = cmap_fun(edge_stylei["color"])

            # These are not the actual edges drawn, only stubs to establish
            # the styles which are then fed into the dynamic, optimised
            # factory (the collection) below
            patch = make_undirected_edge_patch(
                **edge_stylei,
            )
            edgepatches.append(patch)
            adjacent_vertex_ids.append((vid1, vid2))

        if ("cmap" in edge_style) and ("norm" not in edge_style):
            vmin = np.min(colorarray)
            vmax = np.max(colorarray)
            norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
            edge_style["norm"] = norm

        self._edges = EdgeCollection(
            edgepatches,
            vertex_ids=adjacent_vertex_ids,
            vertex_collection=self._vertices,
            labels=labels,
            transform=self.get_offset_transform(),
            style=edge_style,
            directed=self._ipx_internal_data["directed"],
        )
        if "cmap" in edge_style:
            self._edges.set_array(colorarray)

        if self.get_ndim() == 3:
            edge_collection_2d_to_3d(
                self._edges,
            )

    @_stale_wrapper
    def draw(self, renderer):
        """Draw each of the children, with some buffering mechanism."""
        if not self.get_children():
            self._add_vertices()
            self._add_edges()

        if not self.get_visible():
            return

        # Handle zorder manually, just like in AxesBase in mpl
        children = list(self.get_children())
        children.sort(key=lambda x: x.zorder)
        for child in children:
            child.draw(renderer)


def _update_from_internal(style, row, kind):
    """Update single vertex/edge style from internal data."""
    if "color" in row:
        style["color"] = row["color"]
    if "facecolor" in row:
        style["facecolor"] = row["facecolor"]
    if "edgecolor" in row:
        if kind == "vertex":
            style["edgecolor"] = row["edgecolor"]
        else:
            style["color"] = row["edgecolor"]

    if "linewidth" in row:
        style["linewidth"] = row["linewidth"]
    if "linestyle" in row:
        style["linestyle"] = row["linestyle"]
    if "alpha" in row:
        style["alpha"] = row["alpha"]



"""
Train tracks
============

This example demonstrates how to visualise train tracks using ``iplotx``.
The layout, ports, edges, and colors are crafted by hand, as you would
expect for such a bespoke diagram.

"""

import igraph as ig
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import iplotx as ipx


edges = [
    (0, 1),
    (1, 2),
    (2, 3),
    (3, 4),
    (4, 5),
    (6, 7),
    (7, 2),
    (2, 8),
    (8, 9),
    (9, 10),
    (10, 11),
    (12, 13),
    (13, 4),
    (4, 14),
    (14, 15),
]

g = ig.Graph(edges, directed=False)

layout = np.array([
    [0, 0],
    [1, 0],
    [2, 1],
    [2, 3],
    [2, 5],
    [1, 7],
    [5, 1],
    [4, 1],
    [0, 1],
    [0, 2],
    [3, 2],
    [5, 3],
    [4, 6],
    [4, 5],
    [0, 5],
    [-1, 4],
], dtype=float)


# Use edge ports to control the train tracks placement between stations
ports_dict = {
    (1, 2): ("e", "s"),
    (4, 5): ("n", "se"),
    (8, 9): ("w", "w"),
    (10, 11): ("e", "sw"),
    (14, 15): ("w", "n"),
}
ports = [ports_dict.get(edge, None) for edge in edges]
tension = [0.5 * int(port is not None) for port in ports]

# Color stations and tracks by line
edge_color = ["tomato"] * 5 + ["steelblue"] * 6 + ["darkseagreen"] * 4
vertex_color = ["tomato"] * 6 + ["steelblue"] * 6 + ["darkseagreen"] * 4

# Hub stations are black
vertex_color[2] = "black"
vertex_color[4] = "black"

# Visualise
fig, ax = plt.subplots(figsize=(7, 7))
ipx.network(
    g,
    layout=layout,
    ax=ax,
    edge_ports=ports,
    edge_curved=True,
    edge_tension=tension,
    aspect="equal",
    vertex_labels=True,
    edge_linewidth=4,
    vertex_linewidth=4,
    vertex_facecolor="white",
    vertex_label_color="black",
    vertex_edgecolor=vertex_color,
    edge_color=edge_color,
    edge_capstyle="butt",
    vertex_zorder=3,
)
fig.tight_layout()



"""
Directed Acyclic Graph
======================

This example demonstrates how to create a random directed acyclic graph (DAG) and how to use `igraph` layouts based on strings.
"""

import random
import igraph as ig
import iplotx as ipx


# First, we set a random seed for reproducibility.
random.seed(0)

# We generate a random undirected graph with a fixed number of edges, without loops.
g = ig.Graph.Erdos_Renyi(n=15, m=30, directed=False, loops=False)

# Then we convert it to a DAG *in place* and replot it:
g.to_directed(mode="acyclic")

ipx.plot(
    g,
    layout="sugiyama",
    style={
        "vertex": {
            "size": 15,
            "facecolor": "grey",
            "edgecolor": "black",
        },
        "edge": {
            "color": "#222",
            "linewidth": 1,
        },
    },
)



import pytest
import numpy as np
import matplotlib as mpl
from iplotx.edge import ports


def test_port_angles():
    trans_inv = mpl.transforms.IdentityTransform().transform
    for portstring, expected in ports.port_dict.items():
        res = ports._get_port_unit_vector(portstring, trans_inv)
        np.testing.assert_array_almost_equal(res, expected, decimal=6)


def test_port_angles_inverted():
    trans_inv = mpl.transforms.Affine2D().scale(-1, -1).transform
    port_dict = {
        "e": ports.port_dict["w"],
        "n": ports.port_dict["s"],
        "w": ports.port_dict["e"],
        "s": ports.port_dict["n"],
        "nw": ports.port_dict["se"],
        "sw": ports.port_dict["ne"],
        "ne": ports.port_dict["sw"],
        "se": ports.port_dict["nw"],
    }
    for portstring, expected in port_dict.items():
        res = ports._get_port_unit_vector(portstring, trans_inv)
        np.testing.assert_array_almost_equal(res, expected, decimal=6)


def test_port_unsupported():
    trans_inv = mpl.transforms.IdentityTransform().transform
    with pytest.raises(KeyError):
        ports._get_port_unit_vector("unsupported", trans_inv)



"""
Module containing code to manipulate edge visualisations in 3D, especially the Edge3DCollection class.
"""

from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import (
    Line3DCollection,
)

from ...utils.matplotlib import (
    _forwarder,
)
from ...edge import (
    EdgeCollection,
)
from .arrow import (
    arrow_collection_2d_to_3d,
)
from .geometry import (
    _compute_edge_segments as _compute_single_edge_segments,
)


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class Edge3DCollection(Line3DCollection):
    """Collection of vertex patches for plotting."""

    def get_children(self) -> tuple:
        children = []
        if hasattr(self, "_subedges"):
            children.append(self._subedges)
        if hasattr(self, "_arrows"):
            children.append(self._arrows)
        if hasattr(self, "_label_collection"):
            children.append(self._label_collection)
        return tuple(children)

    def set_figure(self, fig) -> None:
        super().set_figure(fig)
        for child in self.get_children():
            child.set_figure(fig)

    @property
    def axes(self):
        return Line3DCollection.axes.__get__(self)

    @axes.setter
    def axes(self, new_axes):
        Line3DCollection.axes.__set__(self, new_axes)
        for child in self.get_children():
            child.axes = new_axes

    _get_adjacent_vertices_info = EdgeCollection._get_adjacent_vertices_info

    def _compute_edge_segments(self):
        """Compute the edge segments for all edges."""
        vinfo = self._get_adjacent_vertices_info()

        segments3d = []
        for vcoord_data in vinfo["offsets"]:
            segment = _compute_single_edge_segments(
                vcoord_data,
            )
            segments3d.append(segment)
        self.set_segments(segments3d)

    def _update_before_draw(self) -> None:
        """Update the collection before drawing."""
        if isinstance(self.axes, Axes3D) and hasattr(self, "do_3d_projection"):
            self.do_3d_projection()

        # TODO: Here's where we would shorten the edges to fit the vertex
        # projections from 3D onto 2D, if we wanted to do that. Because edges
        # in 3D are chains of segments rathen than splines, the shortening
        # needs to be done in a different way to how it's done in 2D.

    def draw(self, renderer) -> None:
        """Draw the collection of vertices in 3D.

        Parameters:
            renderer: The renderer to use for drawing.
        """
        # Prepare the collection for drawing
        self._update_before_draw()

        # Render the Line3DCollection
        # NOTE: we are NOT calling EdgeCollection.draw here
        super().draw(renderer)

        # This sets the labels offsets
        # TODO: implement labels in 3D (one could copy the function from 2D,
        # but would also need to promote the 2D labels into 3D labels similarly to
        # how it's done for 3D vertices).
        # self._update_labels()

        # Now attempt to draw the arrows
        for child in self.get_children():
            child.draw(renderer)


def edge_collection_2d_to_3d(
    col: EdgeCollection,
    zdir: str = "z",
    axlim_clip: bool = False,
):
    """Convert a 2D EdgeCollection to a 3D Edge3DCollection.

    Parameters:
        col: The 2D EdgeCollection to convert.
        zs: The z coordinate(s) to use for the 3D vertices.
        zdir: The axis to use as the z axis (default is "z").
        depthshade: Whether to apply depth shading (default is True).
        axlim_clip: Whether to clip the vertices to the axes limits (default is False).
    """
    if not isinstance(col, EdgeCollection):
        raise TypeError("vertices must be a VertexCollection")

    # NOTE: after this line, none of the EdgeCollection methods will work
    # It's become a static drawer now. It uses segments instead of paths.
    col.__class__ = Edge3DCollection
    col._compute_edge_segments()

    col._axlim_clip = axlim_clip

    # Convert the arrow collection if present
    if hasattr(col, "_arrows"):
        segments3d = col._segments3d

        # Fix the x and y to the center of the target vertex (for now)
        col._arrows._offsets[:] = [segment[-1][:2] for segment in segments3d]
        zs = [segment[-1][2] for segment in segments3d]
        arrow_collection_2d_to_3d(
            col._arrows,
            zs=zs,
            zdir=zdir,
            depthshade=False,
            axlim_clip=axlim_clip,
        )



"""
Module containing leaf edges, i.e. special edges of tree visualisations
that connect leaf vertices to the deepest leaf (typically for labeling).
"""

from typing import (
    Sequence,
    Optional,
    Any,
)
import numpy as np
import pandas as pd
import matplotlib as mpl

from ..utils.matplotlib import (
    _forwarder,
)
from ..vertex import VertexCollection
from iplotx.edge import EdgeCollection


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class LeafEdgeCollection(EdgeCollection):
    """Artist for leaf edges in tree visualisations."""

    def __init__(
        self,
        patches: Sequence[mpl.patches.Patch],
        vertex_leaf_ids: Sequence[tuple],
        vertex_collection: VertexCollection,
        leaf_collection: VertexCollection,
        *args,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        arrow_transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        directed: bool = False,
        style: Optional[dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        """Initialise a LeafEdgeCollection.

        Parameters:
            patches: A sequence (usually, list) of matplotlib `Patch`es describing the edges.
            vertex_ids: A sequence of pairs `(v1, v2)`, each defining the ids of vertices at the
                end of an edge.
            vertex_collection: The VertexCollection instance containing the Artist for the
                vertices. This is needed to compute vertex borders and adjust edges accordingly.
            transform: The matplotlib transform for the edges, usually transData.
            arrow_transform: The matplotlib transform for the arrow patches. This is not the
                *offset_transform* of arrows, which is set equal to the edge transform (previous
                parameter). Instead, it specifies how arrow size scales, similar to vertex size.
                This is usually the identity transform.
            directed: Whether the graph is directed (in which case arrows are drawn, possibly
                with zero size or opacity to obtain an "arrowless" effect).
            style: The edge style (subdictionary: "edge") to use at creation.
        """
        self._leaf_collection = leaf_collection
        super().__init__(
            patches=patches,
            vertex_ids=vertex_leaf_ids,
            vertex_collection=vertex_collection,
            *args,
            transform=transform,
            arrow_transform=arrow_transform,
            directed=directed,
            style=style,
            **kwargs,
        )

    def _get_adjacent_vertices_info(self):
        lindex = self._leaf_collection.get_index()
        lindex = pd.Series(
            np.arange(len(lindex)),
            index=lindex,
        )
        vindex = self._vertex_collection.get_index()
        vindex = pd.Series(
            np.arange(len(vindex)),
            index=vindex,
        ).loc[lindex.index]

        voffsets = []
        vpaths = []
        vsizes = []
        for vid in self._vertex_ids:
            # NOTE: these are in the original layout coordinate system
            # not cartesianised yet.
            offset1 = self._vertex_collection.get_layout().values[vindex[vid]]
            offset2 = self._leaf_collection.get_layout().values[lindex[vid]]
            voffsets.append((offset1, offset2))

            path1 = self._vertex_collection.get_paths()[vindex[vid]]
            path2 = self._leaf_collection.get_paths()[lindex[vid]]
            vpaths.append((path1, path2))

            # NOTE: This needs to be computed here because the
            # VertexCollection._transforms are reset each draw in order to
            # accomodate for DPI changes on the canvas
            size1 = self._vertex_collection.get_sizes_dpi()[vindex[vid]]
            size2 = self._leaf_collection.get_sizes_dpi()[lindex[vid]]
            vsizes.append((size1, size2))

        return {
            "ids": [(vid, vid) for vid in self._vertex_ids],
            "offsets": voffsets,
            "paths": vpaths,
            "sizes": vsizes,
            "loops": False,
        }



"""
scikit-bio tree
===============

This example shows how to use `iplotx` to plot trees from `skbio` or scikit-bio.
"""

from io import StringIO
from skbio import TreeNode
import iplotx as ipx

tree = TreeNode.read(
    StringIO(
        "((),((),(((),()),((),()))));",
    )
)
ipx.plotting.tree(
    tree,
    layout="radial",
    aspect=1,
    edge_color="purple",
)

# %%
# `iplotx` can compute a standard rectangular tree layout as well, and usual style modifications
# work for trees same as networks:

ipx.plotting.tree(
    tree,
    style=[
        "tree",
        {
            "edge": {
                "color": "deeppink",
                "linewidth": 4,
            },
        },
    ],
)



"""
Antibody clones
===============

This example demonstrates how to visualise the antibodies of a large repertoire clone with ``iplotx``.

This particular example uses igraph to load and process the network data, but you can also use networkx, the internal
data structures of ``iplotx``, or any other library you prefer.

Data source: Horns et al., Quake. et al. (2016): https://elifesciences.org/articles/16578.
"""

import json
import igraph as ig
import pandas as pd
import matplotlib.pyplot as plt
import iplotx as ipx


# The original data format is a JSON file with source: {target1: distance1, target2: distance2, ...}
# We convert it into a DataFrame for igraph
with open("data/80201010000000001.mst") as handle:
    data = json.load(handle)
edge_data = {"source": [], "target": [], "distance": []}
for source, target_dict in data.items():
    for target, distance in target_dict.items():
        edge_data["source"].append(source)
        edge_data["target"].append(target)
        edge_data["distance"].append(distance)
edge_data = pd.DataFrame(edge_data)
edge_data["weight"] = 1 / edge_data["distance"]  # Invert distance to get some kind of weight

# NOTE: This particular format is a directed graph, from germline antibody to hypermutated antibody
g = ig.Graph.DataFrame(edge_data, directed=True, use_vids=False)

# Color nodes by distance from the germline antibody
germline = "8031,NA,germline,NA"
depths = {germline: 0.0}
to_visit = [germline]
while to_visit:
    node = to_visit.pop()
    for child, dist in data.get(node, {}).items():
        depths[child] = depths[node] + dist
        to_visit.append(child)
depth_max = max(depths.values())
colors = [depths[name] for name in g.vs["name"]]

# Compute bipartite layout
layout = g.layout_fruchterman_reingold()

fig, ax = plt.subplots(figsize=(8, 7))
artist = ipx.network(
    g,
    layout=layout,
    ax=ax,
    vertex_facecolor=colors,
    vertex_cmap=plt.cm.copper,
    vertex_alpha=0.5,
    vertex_size=5,
    edge_alpha=0.2,
    edge_arrow_width=2,
)[0]
fig.colorbar(
    artist.get_vertices(),
    ax=ax,
    label="Distance from germline\n[# mutations]",
    aspect=10,
    shrink=0.5,
)

# Label the germline antibody for clarity
coords_germline = layout[g.vs["name"].index(germline)]
ax.scatter([coords_germline[0]], [coords_germline[1]], color="tomato", s=80, marker="*", label="Germline")
ax.legend()
fig.tight_layout()


# %%
# This graph turns out to be a tree, so we can revisualise the same data using a tree layout. As an example,
# we use our internal tree data structure, which is a glorified dictionary:

tree = {
    "children": [],
    "name": germline,
    "branch_length": 0.0,
}
to_visit = [(tree, germline)]
while to_visit:
    node, key = to_visit.pop()
    for child, dist in data.get(key, {}).items():
        child_node = {"children": [], "branch_length": dist, "name": child}
        node["children"].append(child_node)
        to_visit.append((child_node, child))
tree = ipx.ingest.providers.tree.simple.SimpleTree.from_dict(tree)

fig, ax = plt.subplots(figsize=(5, 10))
artist = ipx.tree(
    tree,
    ax=ax,
    edge_color="branch_length",
    edge_cmap=plt.cm.plasma,
)
fig.colorbar(
    artist.get_edges(),
    ax=ax,
    label="# mutations\non branch",
    fraction=0.07,
    aspect=10,
    shrink=0.5,
)
fig.tight_layout()



"""
Angular layout and edge waypoints
=================================

This example demonstrates `iplotx`'s capability to style tree appearance with
an angular look, either for the whole tree or specific edges.

.. tip::
    "Waypoints" and "angular" are opposites. When an edge uses waypoints, you
    can say it is using an angular style.

"""

from Bio import Phylo
from io import StringIO
import matplotlib.pyplot as plt
import iplotx as ipx

tree = Phylo.read(
    StringIO("(((A,B),(C,D)),(E,F,G));"),
    format="newick",
)

fig, axs = plt.subplots(2, 2, figsize=(9, 9))
ipx.tree(
    tree,
    ax=axs[0, 0],
    title="Default\n(with waypoints)",
)
ipx.tree(
    tree,
    ax=axs[0, 1],
    layout_angular=True,
    title="Angular",
)
ipx.tree(
    tree,
    ax=axs[1, 0],
    edge_waypoints=False,
    title="Angular\n(via edge_waypoints)",
)
ipx.tree(
    tree,
    ax=axs[1, 1],
    edge_waypoints=[True, False],
    title="Mixed",
)

# %%
# .. warning::
#     For waypoints, ``None`` and the string ``"none"`` are interpreted differently
#     by ``iplotx``. The latter ``"none"`` means no waypoints, whereas the Python
#     singleton ``None`` means letting ``iplotx`` determine the correct type of
#     waypoints. ``None`` is equivalent to ``True``, while ``"none"`` is equivalent
#     to ``False``.
#
# You can even specify angularity on a per-edge basis with a defaultdict:

from collections import defaultdict

ipx.tree(
    tree,
    edge_waypoints=defaultdict(lambda: True, {tree.clade[0, 1]: False}),
    title="Mixed\n(per-edge)",
)

# %%
# .. note::
#   ``iplotx``'s fallback for per-element styling is the constructor for the
#   **type** used by the first value of the dictionary, in this case ``bool()``
#   which returns ``False``. Therefore, if we are ok with angular edges as the
#   fallback, we can skip the defaultdict altogether:

ipx.tree(
    tree,
    edge_waypoints={tree.clade[0, 1]: True},
    title="Mixed per-edge\n(default fallback)",
)

# %%
# You can also mix booleans and other types, although it is not very readable:

ipx.tree(
    tree,
    edge_waypoints=[True, "none"],
    title="Mixed",
)



import pytest
import unittest
import importlib
import matplotlib as mpl

mpl.use("agg")
import matplotlib.pyplot as plt
import iplotx as ipx
from utils import image_comparison

if importlib.util.find_spec("igraph") is None:
    raise unittest.SkipTest("igraph not found, skipping tests")


@pytest.fixture
def graph_and_layout_small():
    import igraph as ig

    g = ig.Graph.Ring(5)
    layout = g.layout("circle").coords
    zs = [0, 1, 0.5, 0.8, 0.2]
    layout3d = [(x, y, z) for (x, y), z in zip(layout, zs)]

    return {
        "graph": g,
        "layout2d": layout,
        "layout3d": layout3d,
    }


@image_comparison(baseline_images=["undirected"], remove_text=True)
def test_undirected(graph_and_layout_small):
    fig = plt.figure(figsize=(6, 6))
    ax = fig.add_subplot(1, 1, 1, projection="3d")

    ipx.network(
        graph_and_layout_small["graph"],
        layout=graph_and_layout_small["layout3d"],
        ax=ax,
    )


@image_comparison(baseline_images=["directed"], remove_text=True)
def test_directed(graph_and_layout_small):
    graph_and_layout_small["graph"].to_directed()
    fig = plt.figure(figsize=(6, 6))
    ax = fig.add_subplot(1, 1, 1, projection="3d")

    ipx.network(
        graph_and_layout_small["graph"],
        layout=graph_and_layout_small["layout3d"],
        ax=ax,
        vertex_alpha=0.3,
    )


@image_comparison(baseline_images=["vertex_labels"], remove_text=True)
def test_vertex_labels(graph_and_layout_small):
    fig = plt.figure(figsize=(6, 6))
    ax = fig.add_subplot(1, 1, 1, projection="3d")

    ipx.network(
        graph_and_layout_small["graph"],
        layout=graph_and_layout_small["layout3d"],
        ax=ax,
        node_marker="s",
        vertex_labels=True,
        edge_zorder=2,
        vertex_zorder=3,
    )



import pytest
import numpy as np
from functools import partial
import matplotlib as mpl
from iplotx.utils import geometry


@pytest.fixture
def three_points():
    return np.array(
        [
            [0, 0],
            [1, 0],
            [1, 1],
        ]
    )


@pytest.fixture
def four_points():
    return np.array(
        [
            [0, 0],
            [1, 0],
            [1, 1],
            [0, 1],
        ]
    )


# 13: two tests fail with AssertionError
def assert_array_equal_upon_cycling(points1, points2):
    """Determine array equality but allow order to be different by cycling.

    Parameters:
        points1: First array of points.
        points2: Second array of points.
    Returns:
        True if the arrays are equal upon cycling, False otherwise.
    """

    def _fun(points1, points2):
        if len(points1) != len(points2):
            return False

        points1 = np.asarray(points1)
        points2 = np.asarray(points2).astype(points1.dtype)

        if (points1.ndim != 1) or (points2.ndim != 1):
            raise ValueError("Only 1D arrays are supported.")

        if (points1 == points2).all():
            return True

        # Cycle
        points2 = np.concatenate([points2, points2])
        i0 = np.flatnonzero((points2 == points1[0]))
        if len(i0) != 2:
            return False
        i0 = i0[0]
        points2 = points2[i0 : i0 + len(points1)]

        # Check again
        return bool((points1 == points2).all())

    assert _fun(points1, points2)


def test_squared_bezier(three_points):
    ev = partial(geometry._evaluate_squared_bezier, three_points)
    np.testing.assert_array_almost_equal(ev(0), three_points[0])
    np.testing.assert_array_almost_equal(ev(1), three_points[2])
    np.testing.assert_array_almost_equal(
        ev(0.5),
        [0.75, 0.25],
    )


def test_cubic_bezier(four_points):
    ev = partial(geometry._evaluate_cubic_bezier, four_points)
    np.testing.assert_array_almost_equal(ev(0), four_points[0])
    np.testing.assert_array_almost_equal(ev(1), four_points[3])
    np.testing.assert_array_almost_equal(
        ev(0.5),
        [0.75, 0.5],
    )


def test_hull_empty():
    hull = geometry.convex_hull([])
    np.testing.assert_array_equal(hull, [])


def test_hull_one(three_points):
    hull = geometry.convex_hull(three_points[:1])
    np.testing.assert_array_equal(hull, [0])


def test_hull_two(three_points):
    hull = geometry.convex_hull(three_points[:2])
    np.testing.assert_array_equal(hull, [0, 1])


def test_hull_three(three_points):
    hull = geometry.convex_hull(three_points)
    assert_array_equal_upon_cycling(hull, [2, 1, 0])


def test_hull_four(four_points):
    hull = geometry.convex_hull(four_points)
    assert_array_equal_upon_cycling(hull, [3, 2, 1, 0])


def test_hull_internal_three(three_points):
    hull = geometry._convex_hull_Graham_scan(three_points)
    # NOTE: It's ok to start from a different point, as layout_small_ring
    # as the subsequent is the same
    np.testing.assert_array_equal(hull, [0, 2, 1])


def test_hull_internal_four(four_points):
    hull = geometry._convex_hull_Graham_scan(four_points)
    # NOTE: It's ok to start from a different point, as layout_small_ring
    # as the subsequent is the same
    np.testing.assert_array_equal(hull, [0, 3, 2, 1])


def test_hull_internal_repetition(four_points):
    points = np.vstack([four_points[:3], four_points[2:]])
    hull = geometry._convex_hull_Graham_scan(points)
    # NOTE: It's ok to start from a different point, as layout_small_ring
    # as the subsequent is the same
    np.testing.assert_array_equal(hull, [0, 4, 3, 1])


def test_hull_internal_repetition_internal(four_points):
    points = np.vstack([four_points[:4], 2 * four_points[2:3], four_points[3:]])
    hull = geometry._convex_hull_Graham_scan(points)
    # NOTE: It's ok to start from a different point, as layout_small_ring
    # as the subsequent is the same
    np.testing.assert_array_equal(hull, [0, 5, 4, 1])


def test_hull_internal_repetition_internal_backtrack():
    points = np.array(
        [
            (0, 0),
            (1, 0),
            (0.5, 0.5),
            # This has a larger angle but created a concave bay that needs backtracking
            (1, 1.1),
            (0, 1),
        ]
    )
    hull = geometry._convex_hull_Graham_scan(points)
    # NOTE: It's ok to start from a different point, as layout_small_ring
    # as the subsequent is the same
    np.testing.assert_array_equal(hull, [0, 4, 3, 1])


def test_hull_internal_repetition_internal_dlast():
    points = np.array(
        [
            (0, 0),
            (1, 0),
            (1, 1),
            # This has a same angle as last but shorter distance, should be ignored
            (0.5, 0.5),
            (0, 1),
        ]
    )
    hull = geometry._convex_hull_Graham_scan(points)
    # NOTE: It's ok to start from a different point, as layout_small_ring
    # as the subsequent is the same
    np.testing.assert_array_equal(hull, [0, 4, 2, 1])


def test_vertexpadding_none(three_points):
    hull_idx = geometry.convex_hull(three_points[:1])
    hull = three_points[hull_idx]
    padding = geometry._compute_group_path_with_vertex_padding(
        hull,
        np.array([three_points[0]] * 31),
        mpl.transforms.IdentityTransform(),
        vertexpadding=0,
    )
    assert len(padding) == 31


def test_vertexpadding_zero():
    padding = geometry._compute_group_path_with_vertex_padding(
        [],
        np.array([]),
        mpl.transforms.IdentityTransform(),
        vertexpadding=0,
    )
    assert padding is None


def test_vertexpadding_one(three_points):
    hull_idx = geometry.convex_hull(three_points[:1])
    hull = three_points[hull_idx]
    padding = geometry._compute_group_path_with_vertex_padding(
        hull,
        np.array([three_points[0]] * 31),
        mpl.transforms.IdentityTransform(),
        vertexpadding=10,
    )
    assert len(padding) == 31


def test_vertexpadding_two(three_points):
    hull_idx = geometry.convex_hull(three_points[:2])
    hull = three_points[hull_idx]
    padding = geometry._compute_group_path_with_vertex_padding(
        hull,
        np.array([three_points[0]] * 61),
        mpl.transforms.IdentityTransform(),
        vertexpadding=10,
    )
    assert len(padding) == 61


def test_vertexpadding_three(three_points):
    hull_idx = geometry.convex_hull(three_points)
    hull = three_points[hull_idx]
    padding = geometry._compute_group_path_with_vertex_padding(
        hull,
        np.array(
            [three_points[0]] * 30
            + [three_points[1]] * 30
            + [three_points[2]] * 30
            + [three_points[0]]
        ),
        mpl.transforms.IdentityTransform(),
        vertexpadding=10,
    )

    # 3 points, each with 30 points + 1 for the closing point
    assert len(padding) == 3 * 30 + 1



"""
Split edges
===========

This example shows how to use split edges, i.e. how to apply different styles to
tree edges that are composed of multiple segments. See also :doc:`plot_simpletreedataprovider`
for the basic example of how to plot a tree with ``iplotx``'s internal tree representation.

.. note::
    The current implementation only supports two-piece edges.
"""

from collections import defaultdict
import matplotlib.pyplot as plt
import iplotx as ipx

tree = {
    "children": (
        {},
        {
            "children": (
                {
                    "children": (
                        {},
                        {}
                    )
                },
                {
                    "children": (
                        {},
                        {}
                    )
                }
            )
        }
    )
}

tree = ipx.ingest.providers.tree.simple.SimpleTree.from_dict(tree)

ipx.tree(
    tree,
    style="tree",
    edge_split_linestyle=":",
    edge_split_color=defaultdict(lambda: "k", {tree.children[1]: "red"}),
)



"""
Feedback loops
==============

Graphs with arrows and bars are commonly used to represent feedback loops in systems.
This example shows how to obtain that effect in ``iplotx``.
"""

import numpy as np
import matplotlib.pyplot as plt
import iplotx as ipx

graph = {
    "edges": [
        ("A", "B"),
        ("B", "C"),
        ("C", "A"),
    ],
    "directed": True,
}
layout = {
    "A": (0, 0),
    "B": (1, 0),
    "C": (0.5, 0.5 * 1.73),
}

fig, ax = plt.subplots()
ipx.network(
    graph,
    layout,
    ax=ax,
    vertex_size=50,
    edge_linewidth=5,
    edge_arrow_width=30,
    edge_arrow_height=40,
    edge_arrow_marker=[")>", "|", "|"],
    edge_shrink=15,
    edge_curved=True,
    edge_tension=[2, 2, 2],
    vertex_facecolor=["#26677FFF", "#635C72FF", "#89374FFF"],
    vertex_edgecolor="black",
    vertex_linewidth=5,
    aspect=1.0,
    margins=0.1,
)

# Add a parametric spiral
t = np.linspace(0, 4 * np.pi, 400)
ax.plot(
    0.5 + np.cos(t + np.pi) * t / 40,
    0.31 + np.sin(t + np.pi) * t / 40,
    color="grey",
    lw=4,
)



import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib import patches
import iplotx as ipx


def vertex_collection(**kwargs):
    if "layout" not in kwargs:
        kwargs["layout"] = pd.DataFrame(
            data=np.array([[0, 0], [1, 1]], np.float64),
            index=["A", "B"],
        )
    return ipx.artists.VertexCollection(**kwargs)


def edge_collection(**kwargs):
    if vertex_collection not in kwargs:
        kwargs["vertex_collection"] = vertex_collection()
    if "vertex_ids" not in kwargs:
        vids = kwargs["vertex_collection"].get_index()
        nedges = 4
        kwargs["vertex_ids"] = [
            [vids[i % len(vids)], vids[(i + 1) % len(vids)]] for i in range(nedges)
        ]
        kwargs["patches"] = [
            patches.PathPatch(mpl.path.Path(np.array([[0, 0], [1, 1]]))),
        ] * len(kwargs["vertex_ids"])

    return ipx.artists.EdgeCollection(**kwargs)


def test_props():
    edges = edge_collection()
    edges.directed = True
    assert edges._directed is True
    assert edges.get_curved() is False


def test_curved():
    edges = edge_collection()
    edges.set_curved(True)


def test_split():
    with ipx.style.context("dashdepth"):
        style = ipx.style.get_style(".edge")
        edge_collection(style=style)


def test_ports():
    edges = edge_collection()
    edges.set_ports(None)
    edges.set_ports(("n", "w"))


def test_tension():
    edges = edge_collection()
    edges.set_tension(None)
    edges.set_tension(0.5)


def test_looptension():
    edges = edge_collection()
    edges.set_looptension(None)
    edges.set_looptension(0.5)


def test_loopmaxangle():
    edges = edge_collection()
    edges.set_loopmaxangle(45)


def test_offset():
    edges = edge_collection()
    edges.set_offset(None)
    edges.set_offset(2)


def test_arrow_marker():
    for marker in [
        "|>",
        "|/",
        "|\\",
        ">",
        "<",
        ">>",
        ")>",
        ")",
        "(",
        "]",
        "[",
        "|",
        "x",
        "s",
        "d",
        "p",
        "q",
    ]:
        with ipx.style.context(edge_arrow_marker=marker):
            edge_collection(directed=True, style=ipx.style.get_style(".edge"))



"""
Breast cancer proteins
======================

This example visualises network of putatively interacting proteins according to the STRING database, filtered to include only entries related to breast cancer.

Vertices are lifted above the edges by increasing their zorder above the edge level, which is 2.
"""

import numpy as np
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx


node_data = pd.read_csv("data/breast_cancer_string_network_coordinates.tsv", sep="\t", index_col=0)
edge_data = pd.read_csv("data/breast_cancer_string_interactions_short.tsv", sep="\t")

g = nx.Graph()
g.add_edges_from(edge_data.iloc[:, :2].values)

layout = node_data[["x_position", "y_position"]]
node_data["color_ipx"] = [list(map(lambda x: float(x) / 256.0, x[4:-1].split(","))) for x in node_data["color"]]
node_data["label_color"] = ["white" if np.mean(x) < 0.4 else "black" for x in node_data["color_ipx"]]


fig, ax = plt.subplots(figsize=(8, 8))
ipx.network(
    g,
    layout=layout,
    ax=ax,
    vertex_labels=True,
    vertex_marker="r",
    vertex_size="label",
    vertex_facecolor=node_data["color_ipx"].to_dict(),
    vertex_label_color=node_data["label_color"].to_dict(),
    vertex_edgecolor="k",
    vertex_linewidth=0.5,
    edge_alpha=0.2,
    vertex_zorder=3,

)
fig.tight_layout()



"""
Cliques
============

This example from `igraph` shows how to compute and visualize cliques and highlight them.

"""

import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx

g = ig.Graph.Famous("Zachary")
layout = g.layout("auto")

cliques = g.cliques(4, 4)

fig, axs = plt.subplots(3, 4)
axs = axs.ravel()
for clique, ax in zip(cliques, axs):
    ipx.plot(
        network=g,
        grouping=ig.VertexCover(g, [clique]),
        layout=layout,
        ax=ax,
        style={
            "vertex": {
                "size": 5,
            },
            "edge": {
                "linewidth": 0.5,
            },
            "grouping": {
                "facecolor": "yellow",
            },
        },
    )
axs[-1].axis("off")  # Hide the last empty subplot


# %%
# Advanced: improving plotting style
# ----------------------------------
# If you want a little more style, you can color the vertices/edges within each
# clique to make them stand out:
fig, axs = plt.subplots(3, 4)
axs = axs.ravel()
for clique, ax in zip(cliques, axs):
    # Color vertices yellow/red based on whether they are in this clique
    g.vs["color"] = "yellow"
    g.vs[clique]["color"] = "red"

    # Color edges black/red based on whether they are in this clique
    clique_edges = g.es.select(_within=clique)
    g.es["color"] = "black"
    clique_edges["color"] = "red"
    # also increase thickness of clique edges
    g.es["width"] = 0.3
    clique_edges["width"] = 1

    ipx.plot(
        network=g,
        grouping=ig.VertexCover(g, [clique]),
        layout=layout,
        ax=ax,
        style={
            "vertex": {
                "size": 5,
                "facecolor": g.vs["color"],
                "edgecolor": "black",
                "linewidth": 0.5,
            },
            "edge": {
                "linewidth": g.es["width"],
                "color": g.es["color"],
            },
            "grouping": {
                "facecolor": "red",
                "vertexpadding": 3,
            },
        },
    )
axs[-1].axis("off")  # Hide the last empty subplot



from typing import (
    Any,
    Optional,
)
import warnings
import numpy as np
import pandas as pd

from .typing import (
    TreeType,
)
from .ingest.typing import (
    TreeDataProvider,
)
import matplotlib as mpl

from .style import (
    copy_with_deep_values,
    rotate_style,
)


class CascadeCollection(mpl.collections.PatchCollection):
    def __init__(
        self,
        tree: TreeType,
        layout: pd.DataFrame,
        layout_name: str,
        orientation: str,
        style: dict[str, Any],
        provider: TreeDataProvider,
        transform: mpl.transforms.Transform,
        maxdepth: Optional[float] = None,
    ):
        self._layout_name = layout_name
        self._orientation = orientation
        style = copy_with_deep_values(style)
        zorder = style.get("zorder", 0)

        # NOTE: there is a weird bug in pandas when using generic Hashable-s
        # with .loc. Seems like doing .T[...] works for individual index
        # elements only though (i.e. using __getitem__ a la dict)
        def get_node_coords(node):
            return layout.T[node].values

        def get_leaves_coords(leaves):
            return np.array(
                [get_node_coords(leaf) for leaf in leaves],
            )

        if "color" in style:
            style["facecolor"] = style["edgecolor"] = style.pop("color")
        extend = style.get("extend", False)

        # These patches need at least a facecolor (usually) or an edgecolor
        # so it's safe to make a list from these
        nodes_unordered = set()
        for prop in ("facecolor", "edgecolor", "linewidth", "linestyle"):
            if prop in style:
                value = style[prop]
                if isinstance(value, dict):
                    nodes_unordered |= set(value.keys())

        if len(nodes_unordered) == 0:
            warnings.warn(
                "No nodes found in the style for the cascading patches. "
                "Please provide a style with at least one dict-like "
                "specification among the following properties: 'facecolor', "
                "'edgecolor', 'color', 'linewidth', or 'linestyle'.",
            )

        # Draw the patches from the closest to the root (earlier drawing)
        # to the closer to the leaves (later drawing).
        drawing_order = []
        for node in provider(tree).preorder():
            if node in nodes_unordered:
                drawing_order.append(node)

        if layout_name not in ("horizontal", "vertical", "radial"):
            raise NotImplementedError(
                f"Cascading patches not implemented for layout: {layout_name}.",
            )

        self._maxdepth = maxdepth

        cascading_patches = []
        nleaves = sum(1 for leaf in provider(tree).get_leaves())
        for node in drawing_order:
            stylei = rotate_style(style, key=node)
            stylei.pop("extend", None)
            # Default alpha is 0.5 for simple colors
            if isinstance(stylei.get("facecolor", None), str) and ("alpha" not in stylei):
                stylei["alpha"] = 0.5

            provider_node = provider(node)
            bl = provider_node.get_branch_length_default_to_one(node)
            node_coords = get_node_coords(node).copy()
            leaves_coords = get_leaves_coords(provider_node.get_leaves())
            if len(leaves_coords) == 0:
                leaves_coords = np.array([node_coords])

            if layout_name in ("horizontal", "vertical"):
                if layout_name == "horizontal":
                    ybot = leaves_coords[:, 1].min() - 0.5
                    ytop = leaves_coords[:, 1].max() + 0.5
                    if orientation == "right":
                        xleft = node_coords[0] - bl
                        xright = maxdepth if extend else leaves_coords[:, 0].max()
                    else:
                        xleft = maxdepth if extend else leaves_coords[:, 0].min()
                        xright = node_coords[0] + bl
                elif layout_name == "vertical":
                    xleft = leaves_coords[:, 0].min() - 0.5
                    xright = leaves_coords[:, 0].max() + 0.5
                    if orientation == "descending":
                        ytop = node_coords[1] + bl
                        ybot = maxdepth if extend else leaves_coords[:, 1].min()
                    else:
                        ytop = maxdepth if extend else leaves_coords[:, 1].max()
                        ybot = node_coords[1] - bl

                patch = mpl.patches.Rectangle(
                    (xleft, ybot),
                    xright - xleft,
                    ytop - ybot,
                    **stylei,
                )
            elif layout_name == "radial":
                dtheta = 2 * np.pi / nleaves
                rmin = node_coords[0] - bl
                rmax = maxdepth if extend else leaves_coords[:, 0].max()
                thetamin = leaves_coords[:, 1].min() - 0.5 * dtheta
                thetamax = leaves_coords[:, 1].max() + 0.5 * dtheta
                thetas = np.linspace(thetamin, thetamax, max(30, (thetamax - thetamin) // 3))
                xs = list(rmin * np.cos(thetas)) + list(rmax * np.cos(thetas[::-1]))
                ys = list(rmin * np.sin(thetas)) + list(rmax * np.sin(thetas[::-1]))
                points = list(zip(xs, ys))
                points.append(points[0])
                codes = ["MOVETO"] + ["LINETO"] * (len(points) - 2) + ["CLOSEPOLY"]

                if "edgecolor" not in stylei:
                    stylei["edgecolor"] = "none"

                path = mpl.path.Path(
                    points,
                    codes=[getattr(mpl.path.Path, code) for code in codes],
                )
                patch = mpl.patches.PathPatch(
                    path,
                    **stylei,
                )

            cascading_patches.append(patch)

        super().__init__(
            cascading_patches,
            transform=transform,
            match_original=True,
            zorder=zorder,
        )

    def get_maxdepth(self) -> float:
        """Get the maxdepth of the cascades.

        Returns: The maximum depth of the cascading patches.
        """
        return self._maxdepth

    def set_maxdepth(self, maxdepth: float):
        """Set the maximum depth of the cascading patches.

        Parameters:
            maxdepth: The new maximum depth for the cascades.

        NOTE: Calling this function updates the cascade patches
        without chechking whether the extent style requires it.
        """
        self._maxdepth = maxdepth
        self._update_maxdepth()

    def _update_maxdepth(self):
        """Update the cascades with a new max depth.

        Note: This function changes the paths without checking whether
        the extent is set or not.
        """
        layout_name = self._layout_name
        orientation = self._orientation

        # This being a PatchCollection, we have to touch the paths
        if layout_name == "radial":
            for path in self.get_paths():
                # Old radii
                r2old = np.linalg.norm(path.vertices[-2])
                # Update the outer part of the wedge patch
                path.vertices[(len(path.vertices) - 1) // 2 :] *= self.get_maxdepth() / r2old
            return

        if (layout_name, orientation) == ("horizontal", "right"):
            for path in self.get_paths():
                path.vertices[[1, 2], 0] = self.get_maxdepth()
        elif (layout_name, orientation) == ("horizontal", "left"):
            for path in self.get_paths():
                path.vertices[[0, 3], 0] = self.get_maxdepth()
        elif (layout_name, orientation) == ("vertical", "descending"):
            for path in self.get_paths():
                path.vertices[[0, 1], 1] = self.get_maxdepth()
        elif (layout_name, orientation) == ("vertical", "ascending"):
            for path in self.get_paths():
                path.vertices[[2, 3], 1] = self.get_maxdepth()
        else:
            raise ValueError(
                f"Layout name and orientation not supported: {layout_name}, {orientation}."
            )



"""
ETE4 tree
=========

This example shows how to use `iplotx` to plot trees from `ete4`.
"""

from ete4 import Tree
import iplotx as ipx

tree = Tree(
    "((),((),(((),()),((),()))));",
)

ipx.plotting.tree(
    tree,
    aspect=1,
    edge_color="grey",
    edge_linestyle=["--", "-"],
)

# %%
# `iplotx` can compute a radial tree layout as well, and usual style modifications
# work for trees same as networks. Moreover, trees have a layout style option to
# choose the starting angle and angular span of the radial layout.

# sphinx_gallery_thumbnail_number = 2
ipx.plotting.tree(
    tree,
    layout="radial",
    layout_orientation="right",
    style=[
        "tree",
        {
            "edge": {
                "color": "deeppink",
                "linewidth": 4,
            },
            "layout": {
                "start": -180,
                "span": 180,
            },
            "leaf": {
                "label": {
                    "hmargin": 15,
                }
            },
            "leafedge": {
                "color": "purple",
                "linewidth": 2,
            },
        },
    ],
    leaf_labels={leaf: str(i + 1) for i, leaf in enumerate(tree.leaves())},
    aspect=1,
    margin=0.1,
)



"""
Module for vertex groupings code, especially the GroupingArtist class.
"""

from typing import Union
import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib.collections import PatchCollection


from .typing import (
    GroupingType,
    LayoutType,
)
from .ingest.heuristics import (
    normalise_layout,
    normalise_grouping,
)
from .style import get_style, rotate_style
from .utils.geometry import (
    convex_hull,
    _compute_group_path_with_vertex_padding,
)


class GroupingArtist(PatchCollection):
    """Matplotlib artist for a vertex grouping (clustering/cover).

    This class is used to plot patches surrounding groups of vertices in a network.
    """

    _factor = 1.0

    def __init__(
        self,
        grouping: GroupingType,
        layout: LayoutType,
        vertexpadding: Union[None, int] = None,
        points_per_curve: int = 30,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        *args,
        **kwargs,
    ) -> None:
        """Container artist for vertex groupings, e.g. covers or clusterings.

        Parameters:
            grouping: This can be a sequence of sets (a la networkx), an igraph Clustering
                or Cover instance (including VertexClustering/VertexCover), or a sequence
                of integers/strings indicating memberships for each vertex.
            layout: The layout of the vertices. If this object has no keys/index, the
                vertices are assumed to have IDs corresponding to integers starting from
                zero.
            vertexpadding: How may points of padding to leave around each vertex centre.
            points_per_curve: How many points to use to approximate a round envelope around
                each convex hull vertex.
            transform: The matplotlib transform to use for the patches (typically transData).
        """
        if vertexpadding is not None:
            self._vertexpadding = vertexpadding
        else:
            style = get_style(".grouping")
            self._vertexpadding = style.get("vertexpadding", 10)

        self._points_per_curve = points_per_curve

        network = kwargs.pop("network", None)
        self.layout = normalise_layout(layout, network=network)
        self.ndim = layout.shape[1]

        patches, grouping, coords_hulls = self._create_patches(
            grouping,
            layout,
            network,
            **kwargs,
        )
        if "network" in kwargs:
            del kwargs["network"]
        self._grouping = grouping
        self._coords_hulls = coords_hulls
        kwargs["match_original"] = True

        super().__init__(patches, *args, **kwargs)

        zorder = get_style(".grouping").get("zorder", 1)
        self.set_zorder(zorder)

        self.set_transform(transform)

    def set_figure(self, figure) -> None:
        """Set the figure for the grouping, recomputing the paths depending on the figure's dpi."""
        ret = super().set_figure(figure)
        self._compute_paths(self.get_figure(root=True).dpi)
        return ret

    @property
    def axes(self):
        return PatchCollection.axes.__get__(self)

    @axes.setter
    def axes(self, new_axes):
        PatchCollection.axes.__set__(self, new_axes)
        for child in self.get_children():
            child.axes = new_axes
        self.set_figure(new_axes.figure)

    def get_layout(self) -> pd.DataFrame:
        """Get the layout used for this grouping."""
        return self.layout

    def get_vertexpadding(self) -> float:
        """Get the vertex padding of each group."""
        return self._vertexpadding

    def get_vertexpadding_dpi(self, dpi: float = 72.0) -> float:
        """Get vertex padding of each group, scaled by dpi of the figure."""
        return self.get_vertexpadding() * dpi / 72.0 * self._factor

    def _create_patches(self, grouping, layout, network, **kwargs):
        grouping = normalise_grouping(grouping, layout)
        style = get_style(".grouping")
        style.pop("vertexpadding", None)

        style.update(kwargs)

        patches = []
        coords_hulls = []
        for i, (name, vids) in enumerate(grouping.items()):
            if len(vids) == 0:
                continue
            vids = np.array(list(vids))
            coords = layout.loc[vids].values
            idx_hull = convex_hull(coords)
            coords_hull = coords[idx_hull]
            coords_hulls.append(coords_hull)

            stylei = rotate_style(style, i)

            # NOTE: the transform is set later on
            patch = _compute_group_patch_stub(
                coords_hull,
                self._vertexpadding,
                label=name,
                **stylei,
            )

            patches.append(patch)
        return patches, grouping, coords_hulls

    def _compute_paths(self, dpi: float = 72.0) -> None:
        for i, hull in enumerate(self._coords_hulls):
            _compute_group_path_with_vertex_padding(
                hull,
                self._paths[i].vertices,
                self.get_transform(),
                vertexpadding=self.get_vertexpadding_dpi(dpi),
            )

    def _process(self) -> None:
        self._compute_paths()

    def draw(self, renderer) -> None:
        """Draw or re-draw the grouping patches.

        Parameters:
            renderer: The renderer to use for drawing the patches.
        """
        # FIXME: this kind of breaks everything since the vertices' magical "_transforms" does
        # not really scale from 72 pixels but rather from the screen's or something.
        # Conclusion: using this keeps consistency across dpis but breaks proportionality of
        # vertexpadding and vertex_size (for now).
        # NOTE: this might be less bad than initially thought in the sense that even perfect
        # scaling does not seem to align the center of the perimeter of the group with the
        # center of the perimeter of the vertex when of the same exact size. So we are
        # probably ok winging it as users will adapt.
        self._compute_paths(self.get_figure(root=True).dpi)
        super().draw(renderer)


def _compute_group_patch_stub(
    points,
    vertexpadding,
    **kwargs,
):
    if vertexpadding == 0:
        return mpl.patches.Polygon(
            points,
            **kwargs,
        )

    # NOTE: Closing point: mpl is a bit quirky here
    vertices = np.zeros(
        (1 + 30 * len(points), 2),
    )
    codes = ["MOVETO"] + ["LINETO"] * (len(vertices) - 2) + ["CLOSEPOLY"]
    codes = [getattr(mpl.path.Path, x) for x in codes]
    patch = mpl.patches.PathPatch(
        mpl.path.Path(
            vertices,
            codes=codes,
        ),
        **kwargs,
    )

    return patch



from typing import Optional, Any
from functools import wraps, partial
from math import atan2
import numpy as np
import pandas as pd
import matplotlib as mpl

from .geometry import (
    _evaluate_squared_bezier,
    _evaluate_cubic_bezier,
)


# NOTE: https://github.com/networkx/grave/blob/main/grave/grave.py
def _stale_wrapper(func):
    """Decorator to manage artist state."""

    @wraps(func)
    def inner(self, *args, **kwargs):
        try:
            func(self, *args, **kwargs)
        finally:
            self.stale = False

    return inner


def _forwarder(forwards, cls=None):
    """Decorator to forward specific methods to Artist children."""
    if cls is None:
        return partial(_forwarder, forwards)

    def make_forward(name):
        def method(self, *args, **kwargs):
            """Each decorated method is called on the decorated class and then, nonrecursively, on all children."""
            ret = getattr(cls.mro()[1], name)(self, *args, **kwargs)
            for c in self.get_children():
                getattr(c, name)(*args, **kwargs)
            return ret

        return method

    for f in forwards:
        method = make_forward(f)
        method.__name__ = f
        method.__doc__ = "broadcasts {} to children".format(f)
        setattr(cls, f, method)

    return cls


def _additional_set_methods(attributes, cls=None):
    """Decorator to add specific set methods for children properties.

    This is useful to autogenerate methods a la set_<key>(value), for
    instance set_alpha(value). It works by delegating to set(alpha=value).

    Overall, this is a minor tweak compared to the previous decorator.
    """
    if cls is None:
        return partial(_additional_set_methods, attributes)

    def make_setter(name):
        def method(self, value):
            self.set(**{name: value})

        return method

    for attr in attributes:
        desc = attr.replace("_", " ")
        method = make_setter(attr)
        method.__name__ = f"set_{attr}"
        method.__doc__ = f"Set {desc}."
        setattr(cls, f"set_{attr}", method)

    return cls


def _get_label_width_height(text, hpadding=18, vpadding=12, dpi=72.0, **kwargs):
    """Get the bounding box size for a text with certain properties.

    Parameters:
        text: The text to measure.
        hpadding: Horizontal padding to add to the width.
        vpadding: Vertical padding to add to the height.
        **kwargs: Additional keyword arguments for text properties. "fontsize" is accepted,
            as is "size". Many other properties are not used and will raise and exception.

    Returns:
        A tuple (width, height) representing the size of the text bounding box. Because
        some text properties such as weight are not taken into account, ths function is not
        very accurate. Yet, it is often good enough and easier to implement than a careful
        orchestration of Figure.draw_without_rendering.
    """
    if len(text) == 0:
        return (0, 0)

    if "fontsize" in kwargs:
        kwargs["size"] = kwargs.pop("fontsize")
    forbidden_props = [
        "horizontalalignment",
        "verticalalignment",
        "ha",
        "va",
        "color",
        "edgecolor",
        "facecolor",
    ]
    for prop in forbidden_props:
        if prop in kwargs:
            del kwargs[prop]

    path = mpl.textpath.TextPath((0, 0), text, **kwargs)
    boundingbox = path.get_extents()
    width = boundingbox.width
    height = boundingbox.height

    # Scaling with font size appears broken... try to patch it up linearly here, even though we
    # know it does not work terribly accurately
    width *= kwargs.get("size", 12) / 12.0
    height *= kwargs.get("size", 12) / 12.0

    width += hpadding
    height += vpadding

    # Scale by dpi
    width *= dpi / 72.0
    height *= dpi / 72.0

    return (width, height)


def _compute_mid_coord_and_rot(path, trans):
    """Compute mid point of an edge, straight or curved."""
    # Straight path
    if path.codes[-1] == mpl.path.Path.LINETO:
        coord = path.vertices.mean(axis=0)
        v1 = path.vertices[0]
        v2 = path.vertices[-1]

    # Cubic Bezier
    elif path.codes[-1] == mpl.path.Path.CURVE4:
        coord = _evaluate_cubic_bezier(path.vertices, 0.5)
        v1 = _evaluate_cubic_bezier(path.vertices, 0.475)
        v2 = _evaluate_cubic_bezier(path.vertices, 0.525)

    # Square Bezier
    elif path.codes[-1] == mpl.path.Path.CURVE3:
        coord = _evaluate_squared_bezier(path.vertices, 0.5)
        v1 = _evaluate_squared_bezier(path.vertices, 0.475)
        v2 = _evaluate_squared_bezier(path.vertices, 0.525)

    else:
        raise ValueError(
            "Curve type not straight and not squared/cubic Bezier, cannot compute mid point."
        )

    v1 = trans(v1)
    v2 = trans(v2)
    rot = atan2(
        v2[1] - v1[1],
        v2[0] - v1[0],
    )
    return coord, rot


def _build_cmap_fun(
    style: dict[str, Any],
    key: str,
    norm=None,
    internal: Optional[pd.DataFrame] = None,
):
    """Map colormap on top of numerical values.

    Parameters:
        style: A dictionary of style properties.
        key: The key in the style dictionary to look for values. Values can be a list/array,
            a dictionary of numerical values, or a string, in which case the corresponding
            column in the "internal" DataFrame is used as an array of numerical values.
        norm: An optional matplotlib Normalize instance. If None, the values are normalized.
        internal: An optional DataFrame, required if "values" is a string.
    """
    values = style[key]
    cmap = style["cmap"]

    cmap = mpl.cm._ensure_cmap(cmap)

    if isinstance(values, str):
        if not isinstance(internal, pd.DataFrame):
            raise ValueError("If 'values' is a string, 'internal' must be a DataFrame.")
        values = internal[values].values
        internal[key] = values

    else:
        if np.isscalar(values):
            values = [values]

        if isinstance(values, dict):
            values = np.array(list(values.values()))

    if norm is None:
        vmin = np.nanmin(values)
        vmax = np.nanmax(values)
        norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    return lambda x: cmap(norm(x))



import pytest
import numpy as np
import matplotlib as mpl
import iplotx as ipx


def test_additional_set_methods_noclass():
    attributes = ["xlim", "ylim"]
    res = ipx.utils.matplotlib._additional_set_methods(attributes)
    assert hasattr(res, "__call__")


def test_additional_set_methods_class():
    attributes = ["xlim", "ylim"]

    class DummyArtist:
        pass

    res = ipx.utils.matplotlib._additional_set_methods(attributes, DummyArtist)
    assert hasattr(res, "set_xlim")
    assert hasattr(res, "set_ylim")


def test_midpoint_squared_bezier():
    path = mpl.path.Path(
        vertices=[[0, 0], [1, 0], [1, 1]],
        codes=[mpl.path.Path.MOVETO, mpl.path.Path.CURVE3, mpl.path.Path.CURVE3],
    )
    coord, rot = ipx.utils.matplotlib._compute_mid_coord_and_rot(
        path,
        mpl.transforms.IdentityTransform().transform,
    )
    assert np.allclose(coord, [0.75, 0.25])
    assert np.isclose(rot, np.pi / 4)


def test_midpoint_invalid():
    path = mpl.path.Path(
        vertices=[[0, 0], [1, 0]],
        codes=[mpl.path.Path.MOVETO, mpl.path.Path.MOVETO],
    )
    with pytest.raises(ValueError):
        ipx.utils.matplotlib._compute_mid_coord_and_rot(
            path,
            mpl.transforms.IdentityTransform().transform,
        )


def test_label_size_emptystring():
    width, height = ipx.utils.matplotlib._get_label_width_height("")
    assert width == 0
    assert height == 0


def test_build_cmap_fun():
    res = ipx.utils.matplotlib._build_cmap_fun(
        {"color": 0, "cmap": "Greys"},
        "color",
    )
    assert np.allclose(res(0), np.ones(4))



"""
Edge tension
============

This example shows how to use edge tension to create curved edges with sharper turns.
"""

import networkx as nx
import iplotx as ipx

tension = [0, 0.1, 0.5, 1.0, 2.0, 4.0, 8.0]
G = nx.MultiGraph()
G.add_edges_from([(0, 1)] * len(tension))

layout = {
    0: (0, 0),
    1: (1, 1),
}

ipx.network(
    G,
    layout=layout,
    edge_curved=True,
    edge_tension=tension,
    edge_labels=[str(x) if x > 0 else "" for x in tension],
    edge_label_bbox=dict(
        edgecolor="black",
        facecolor="white",
        boxstyle="round,pad=0.2",
    ),
    edge_label_rotate=False,
)

# %%
# Negative tension can be used too:

ipx.network(
    G,
    layout=layout,
    edge_curved=True,
    edge_tension=[-x for x in tension],
    edge_labels=[str(-x) if x > 0 else "" for x in tension],
    edge_label_bbox=dict(
        edgecolor="black",
        facecolor="white",
        boxstyle="round,pad=0.2",
    ),
    edge_label_rotate=False,
)

# %%
# For directed graphs, positive tension is always **to the right** when facing out in the edge direction

tension = [1.0, 3.0]
Gd = nx.MultiDiGraph()
Gd.add_edges_from([(0, 1)] * len(tension))
Gd.add_edges_from([(1, 0)] * len(tension))


ipx.network(
    Gd,
    layout=layout,
    edge_curved=True,
    edge_tension=tension * 2,
    edge_labels=[f"+{x}" if x > 0 else "" for x in tension] * 2,
    edge_label_bbox=dict(
        edgecolor="black",
        facecolor="white",
        boxstyle="round,pad=0.2",
    ),
    edge_label_rotate=False,
    aspect=1,
)

# %%
# .. tip::
#   For undirected graphs, the chirality (right/leftness) of the tension depends on whether the
#   network library reorders the edge adjacent vertices (e.g. `networkx` does). Some trial and error on
#   a specific plot should lead to the desired solution.



"""
Foraging table
==============

This example visualises a table of foraging strategies, as found at https://doi.org/10.1111/1365-2656.13609.

.. warning::
  The edge thicknesses are manually set to mimic the original figure a little, not derived from real data. This is just for illustration.
"""

from collections import defaultdict
import numpy as np
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx


common_style = {
    "vertex": {
        "linewidth": 2,
        "edgecolor": "black",
    },
    "edge": {
        "color": "grey",
        "tension": 0.75,
        "curved": True,
        "arrow": {
            "width": 0,
        }
    }
}


# Create figure and subplots
fig, axs = plt.subplots(3, 2, figsize=(6, 9))

# Two examples with 5 nodes
g = nx.complete_graph(range(1, 6), nx.DiGraph)
layout = nx.circular_layout(g)
ipx.plot(
    g, layout, ax=axs[0, 0],
    vertex_labels=True,
    style=common_style,
    vertex_facecolor=["black"] + ["white"] * 4,
    vertex_label_color=["white"] + ["black"] * 4,
    vertex_marker=["o"] * 4 + ["s"],
    vertex_size=[20] * 4 + [25],
    edge_linewidth=defaultdict(lambda: 1, {(4, 5): 1.5, (5, 1): 1.5}),
)
ipx.plot(
    g, layout, ax=axs[0, 1],
    vertex_labels=True,
    style=common_style,
    vertex_facecolor=["black"] + ["white"] * 4,
    vertex_label_color=["white"] + ["black"] * 4,
    vertex_marker=["o"] * 4 + ["s"],
    vertex_size=[25] * 4 + [30],
    edge_linewidth=defaultdict(
        lambda: 3, {
            (4, 2): 1, (4, 3): 1.5, (4, 5): 1.5,
        },
    ),
)

# Two examples with 6 nodes
g = nx.complete_graph([1, 2, 3, 4, 6, 5], nx.DiGraph)
layout = nx.circular_layout(g)
ipx.plot(
    g, layout, ax=axs[1, 0],
    vertex_labels=True,
    style=common_style,
    vertex_facecolor=["black"] + ["white"] * 2 + ["black"] * 3,
    vertex_label_color=["white"] + ["black"] * 2 + ["white"] * 3,
    vertex_size=[30] * 2 + [25, 20]  * 2,
    edge_linewidth=defaultdict(
        lambda: 1, {
            (1, 2): 6, (2, 1): 6, (6, 2): 4, (6, 4): 4, (6, 1): 4, (6, 5): 4, (5, 6): 4, (5, 3): 4, (5, 4): 3,
        },
    ),
)
ipx.plot(
    g, layout, ax=axs[1, 1],
    vertex_labels=True,
    style=common_style,
    vertex_facecolor=["black"] + ["white"] * 2 + ["black"] * 3,
    vertex_label_color=["white"] + ["black"] * 2 + ["white"] * 3,
    vertex_size=[30] * 2 + [20] + [30] * 2 + [25],
    edge_linewidth=defaultdict(
        lambda: 1, {
            (1, 2): 7, (2, 1): 7, (6, 2): 4, (6, 4): 6, (4, 6): 6, (6, 1): 4, (6, 5): 4, (5, 6): 4, (5, 3): 4, (5, 4): 3,
        },
    ),
)

# Two examples with 4 nodes
g = nx.complete_graph(range(1, 5), nx.DiGraph)
layout = nx.circular_layout(g)
ipx.plot(
    g, layout, ax=axs[2, 0],
    vertex_labels=True,
    style=common_style,
    vertex_facecolor=["white"] + ["black"] * 3,
    vertex_label_color=["black"] + ["white"] * 3,
    vertex_size=[25] * 3 + [30],
)

g = nx.complete_graph([1, 2, 3, 5], nx.DiGraph)
layout = nx.circular_layout(g)
ipx.plot(
    g, layout, ax=axs[2, 1],
    vertex_labels=True,
    style=common_style,
    vertex_facecolor=["black"] + ["white"] * 3,
    vertex_label_color=["white"] + ["black"] * 3,
    vertex_size=[20, 25] * 2,
    vertex_marker=["o"] * 3 + ["s"],
    edge_linewidth=defaultdict(
        lambda: 3, {
            (1, 3): 1, (3, 1): 1.5, (3, 5): 1,
        },
    ),
)


fig.tight_layout()



from copy import deepcopy
from collections import defaultdict
import pytest

import iplotx as ipx


def test_reset():
    ipx.style.use("hollow")
    ipx.style.reset()
    assert ipx.style.current == ipx.style.styles["default"]


def test_get_nonexistent_style():
    with pytest.raises(TypeError):
        ipx.style.get_style(".vertex.size.")


def test_get_style_invalid():
    with pytest.raises(ValueError):
        ipx.style.get_style("default", {}, "third_arg")


def test_get_style_args():
    assert ipx.style.get_style(".vertex.curved", 80) == 80


def test_vertex_node():
    style = ipx.style.merge_styles(
        [
            dict(vertex_size=80),
            dict(node_label_color="grey"),
        ]
    )
    assert isinstance(style, dict)
    assert "vertex" in style
    assert "node" not in style
    assert "size" in style.get("vertex", {})
    assert "label" in style.get("vertex", {})
    assert "color" in style.get("vertex", {"label": {}}).get("label", {})


def test_edgepadding():
    style = ipx.style.merge_styles(
        [
            {
                "edge": {
                    "padding": 8,
                    "shrink": 2,
                }
            },
        ]
    )
    assert isinstance(style, dict)
    assert "edge" in style
    assert "shrink" in style.get("edge", {})
    assert "padding" not in style.get("edge", {})

    style = ipx.style.merge_styles(
        [
            {
                "edge": {
                    "padding": 8,
                }
            },
        ]
    )
    assert isinstance(style, dict)
    assert "edge" in style
    assert "shrink" in style.get("edge", {})
    assert "padding" not in style.get("edge", {})


def test_rotate_type_fallback():
    style = {
        "size": {"hello": 80},
    }
    assert ipx.style.rotate_style(
        style,
        key="world",
    ) == {"size": 0}


def test_rotate_style_key2():
    style = {
        "size": {"hello": 80},
    }
    assert ipx.style.rotate_style(
        style,
        key="world",
        key2="hello",
    ) == {"size": 80}


def test_rotate_style_noargs():
    style = {
        "size": {"hello": 80},
    }
    with pytest.raises(ValueError):
        ipx.style.rotate_style(style)


def test_use_style_invalid():
    with pytest.raises(TypeError):
        ipx.style.use(["hollow", ".vertex.size."])
    assert ipx.style.current == ipx.style.styles["default"]


def test_flat_style():
    with ipx.style.context(
        dict(
            vertex_size=80,
            edge_label_bbox_facecolor="yellow",
            zorder=80,
        ),
    ):
        current = ipx.style.current
        assert current["vertex"]["size"] == 80
        assert current["edge"]["label"]["bbox"]["facecolor"] == "yellow"

        with ipx.style.context(
            vertex_size=70,
        ):
            assert current["vertex"]["size"] == 70
            assert current["edge"]["label"]["bbox"]["facecolor"] == "yellow"


def test_generator():
    styles = iter([{"vertex_size": 80}, {"vertex_size": 70}])
    with ipx.style.context(
        styles,
    ):
        assert ipx.style.current["vertex"]["size"] == 70


def test_use():
    style = deepcopy(ipx.style.current)
    ipx.style.use("hollow")
    ipx.style.use("default")
    assert style == ipx.style.current


def test_copy_with_deep_values():
    partial_style = defaultdict(lambda: 80, {"a": 10})
    partial_style_copy = ipx.style.copy_with_deep_values(partial_style)
    assert partial_style_copy["a"] == 10
    assert partial_style_copy["b"] == 80
    assert hasattr(partial_style_copy, "default_factory")

    style = {"vertex": {"size": defaultdict(lambda: 80, {"a": 10})}}
    style_copy = ipx.style.copy_with_deep_values(style)
    assert style_copy["vertex"]["size"]["a"] == 10
    assert style_copy["vertex"]["size"]["b"] == 80
    assert not hasattr(style_copy, "default_factory")
    assert hasattr(style_copy["vertex"]["size"], "default_factory")



"""
Editing artists after plotting
==============================

One of the strengths of `iplotx` is the support for editing artists after
they have been plotted the first time without making a whole new figure.

In this example, we will change vertices, edges, arrows, etc. and check
out how they look like.
"""

# %%
# This is the vanilla version, taken from the "Ports" gallery example:

import igraph as ig
import iplotx as ipx

g = ig.Graph.Ring(3, directed=True)

artist = ipx.network(
    g,
    layout="circle",
    edge_curved=True,
    edge_ports=[
        ("n", "w"),  # exit from the top, enter from the left
        ("e", "s"),  # exit from the right, enter from the bottom
        ("n", "s"),  # exit from the top, enter from the bottom
    ],
    edge_tension=[1.5, 1.8, 0.8],
    edge_color=["tomato", "steelblue", "purple"],
)[0]

# %%
# Now we can change the vertex sizes:

# sphinx_gallery_thumbnail_number = 2
# SAME BLOCK AS BEFORE, ONLY NEEDED TO MAKE A NEW FIGURE
artist = ipx.network(
    g,
    layout="circle",
    edge_curved=True,
    edge_ports=[
        ("n", "w"),  # exit from the top, enter from the left
        ("e", "s"),  # exit from the right, enter from the bottom
        ("n", "s"),  # exit from the top, enter from the bottom
    ],
    edge_tension=[1.5, 1.8, 0.8],
    edge_color=["tomato", "steelblue", "purple"],
)[0]
# END OF SAME BLOCK

artist.get_vertices().set_sizes([5, 5, 40])
artist.get_edges().set_ports(
    [
        ("n", "n"),
        ("w", None),
        ("e", "s"),
    ]
)
artist.get_edges().set_tension([1.3, 0.3, 0.5])



"""
Traveling salesman
==================

This example from networkx demonstrates simple styling of nodes and edges.
"""

import matplotlib.pyplot as plt
import networkx as nx
import networkx.algorithms.approximation as nx_app
import math
import iplotx as ipx


G = nx.random_geometric_graph(20, radius=0.4, seed=3)
pos = nx.get_node_attributes(G, "pos")

# Depot should be at (0,0)
pos[0] = (0.5, 0.5)

H = G.copy()


# Calculating the distances between the nodes as edge's weight.
for i in range(len(pos)):
    for j in range(i + 1, len(pos)):
        dist = math.hypot(pos[i][0] - pos[j][0], pos[i][1] - pos[j][1])
        dist = dist
        G.add_edge(i, j, weight=dist)

cycle = nx_app.christofides(G, weight="weight")
edge_list = list(nx.utils.pairwise(cycle))
edge_color = {
    tuple(e): "red" if e in edge_list or e[::-1] in edge_list else "none" for e in G.edges()
}
nx.set_edge_attributes(G, edge_color, "color")

ipx.plot(
    H,
    layout=pos,
    style={
        "edge": {
            "color": "blue",
            "linewidth": 1,
        },
        "vertex": {
            "facecolor": "none",
            "edgecolor": "none",
        },
    },
)
ipx.plot(
    G,
    ax=plt.gca(),
    layout=pos,
    vertex_labels=True,
    style={
        "edge": {
            "color": G.edges.data("color"),
            "linewidth": 3,
        },
        "vertex": {
            "size": 20,
            "facecolor": "steelblue",
            "edgecolor": "none",
            "label": {
                "color": "black",
            },
        },
    },
)

print("The route of the traveller is:", cycle)



"""
Food network
============

This example visualises a food network, with data taken from https://doi.org/10.3389/fevo.2020.588430 (Supplementary Table S1).
"""

from collections import defaultdict
import numpy as np
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx


# Read in the data on how many times scat evidenced was found that each animal (column) ate each food item (row)
table = pd.read_csv("data/fevo-08-588430_DataSheet1_S1.csv", skiprows=2, usecols=[0, 1, 2, 3, 6, 9, 12, 15, 18, 21]).iloc[1:]


# Only part of the table is relevant to build the network
adjacency_matrix = table.set_index("OTUs").iloc[:, -6:].astype(int)
# Clean up column names a bit
adjacency_matrix.columns = [x.split(" (")[0] for x in adjacency_matrix.columns]

# Convert to a directed edge list
edge_data = adjacency_matrix.T.stack()
edge_data = edge_data[edge_data > 0].reset_index()
edge_data.columns = ["predator", "prey", "weight"]

# Build graph
g = nx.DiGraph()
g.add_weighted_edges_from(edge_data.values)

# Compute force-directed layout
layout = nx.spring_layout(g, seed=42)
layout.update({
    "Bobcat": [0.4, 0.4],
    "Jackrabbit": [-0.4, -0.4],
    "Gray Fox": [-0.4, 0.4],
    "Puma": [0.4, -0.4],
    "Deer": [0.2, 0.0],
    "Coyote": [-0.2, 0.0],
})

# Compute vertex labels and sizes
vertex_labels = {}
vertex_sizes = {}
for node in g.nodes():
    n_preys = g.out_degree(node)
    if n_preys == 0:
        vertex_labels[node] = ""
        vertex_sizes[node] = 15
    else:
        vertex_labels[node] = node
        vertex_sizes[node] = 55

legend_colord = {
    "Herb": "lawngreen",
    "Woody": "green",
    "Grass": "limegreen",
    "Small Mammal": "khaki",
    "Large Mammal": "saddlebrown",
    "Predator": "tomato",
}
vertex_facecolor = table.set_index("OTUs")["Functional Group"].map(legend_colord).to_dict()
# Add predators
for predator in adjacency_matrix.columns:
    vertex_facecolor[predator] =  legend_colord["Predator"]

edge_color = {key: vertex_facecolor[key[1]] for key in g.edges()}
edge_linewidth = {(u, v): 0.5 * z['weight'] for u, v, z in g.edges(data=True)}


fig, ax = plt.subplots(figsize=(8, 6.6))
ipx.network(
    g,
    layout=layout,
    ax=ax,
    vertex_labels=vertex_labels,
    vertex_size=vertex_sizes,
    vertex_facecolor=vertex_facecolor,
    vertex_edgecolor="black",
    vertex_linewidth=1,
    vertex_label_color="black",
    vertex_zorder=3,
    edge_curved=True,
    edge_tension=2,
    edge_alpha=0.7,
    edge_linewidth=edge_linewidth,
    edge_color=edge_color,
)
ax.legend(
    [plt.Circle((0, 0), 0, facecolor=color, edgecolor="black") for color in legend_colord.values()],
    list(legend_colord.keys()),
    loc="lower center",
    ncol=6,
    frameon=False,
)
fig.tight_layout()

# %%
# We can also visualise each predator individually to get a sense of their differences, using subplots:

fig, axs = plt.subplots(3, 2, figsize=(8, 12))
axs = axs.ravel()
for predator, ax in zip(adjacency_matrix.columns, axs):
    subgraph = g.subgraph([predator] + list(g.successors(predator)))
    sublayout = {k: v for k, v in layout.items() if k in subgraph.nodes()}
    sub_vertex_sizes = {k: 0.5 * v for k, v in vertex_sizes.items() if k in subgraph.nodes()}

    ipx.network(
        subgraph,
        layout=sublayout,
        ax=ax,
        vertex_labels=False,
        vertex_size=sub_vertex_sizes,
        vertex_facecolor=vertex_facecolor,
        vertex_edgecolor="black",
        vertex_linewidth=1,
        vertex_zorder=3,
        edge_curved=True,
        edge_tension=2,
        edge_alpha=0.7,
        edge_color=edge_color,
        edge_linewidth=edge_linewidth,
        title=predator,
    )



"""
iplotx is a library for interactive plotting of networks and trees in
matplotlib.

It guarantees the visualisation will look exactly the same no matter what
library was used to construct the network.
"""

from .version import __version__
from .plotting import (
    network,
    graph,
    tree,
    plot,
)
import iplotx.artists as artists
import iplotx.style as style


__all__ = [
    "network",
    "tree",
    "plot",
    "graph",
    "artists",
    "style",
    "__version__",
]



"""
Zero-dependency networks
========================

This example uses `iplotx`'s internal `SimpleNetworkDataProvider` to show how to visualise
networkx without using any external analysis library (e.g. `igraph`, `networkx`).
"""

import iplotx as ipx

network = {
    "edges": [
        ("alice", "bob"),
        ("bob", "jago"),
        ("alice", "jago"),
    ]
}
layout = {
    "alice": (0, 0),
    "bob": (1, 1),
    "jago": (1, 0),
}
ipx.plot(
    network,
    layout,
    vertex_labels=True,
    style="hollow",
)



from io import StringIO
import pytest
import unittest
import importlib
import matplotlib as mpl
import iplotx as ipx

if importlib.util.find_spec("Bio") is None:
    raise unittest.SkipTest("biopython not found, skipping tests")
else:
    from Bio import Phylo


CascadeCollection = ipx.artists.CascadeCollection


def make_small_tree(layout="horizontal"):
    tree = next(
        Phylo.NewickIO.parse(
            StringIO(
                "(()(()((()())(()()))))",
            )
        )
    )

    internal_data = ipx.ingest.ingest_tree_data(
        tree,
        layout,
        directed=False,
        layout_style={},
    )
    layout = internal_data["vertex_df"]
    provider = ipx.ingest.data_providers["tree"][internal_data["tree_library"]]

    return {
        "tree": tree,
        "layout": layout,
        "provider": provider,
    }


def make_cascade(layout="horizontal", orientation="right", style={}):
    small_tree = make_small_tree(layout=layout)
    with ipx.style.context(style):
        style = ipx.style.get_style(".cascade")
        style["color"] = {
            small_tree["tree"].clade[0, 1]: "blue",
        }
        return CascadeCollection(
            tree=small_tree["tree"],
            layout=small_tree["layout"],
            layout_name=layout,
            orientation=orientation,
            style=style,
            provider=small_tree["provider"],
            transform=mpl.transforms.IdentityTransform(),
            maxdepth=5,
        )


def test_init():
    cascade = make_cascade()
    assert isinstance(cascade, CascadeCollection)


def test_init_unsupported():
    with pytest.raises(ValueError):
        make_cascade(layout="unsupported")


def test_update_maxdepth():
    for layout, orientation in [
        ("horizontal", "right"),
        ("horizontal", "left"),
        ("vertical", "ascending"),
        ("vertical", "descending"),
        ("radial", "clockwise"),
        ("radial", "counterclockwise"),
    ]:
        cascade = make_cascade(layout=layout, orientation=orientation)
        cascade._update_maxdepth()


def test_update_maxdepth_unsupported():
    cascade = make_cascade(layout="horizontal", orientation="unsupported")
    with pytest.raises(ValueError):
        cascade._update_maxdepth()


def test_extend_true():
    for layout, orientation in [
        ("horizontal", "right"),
        ("horizontal", "left"),
        ("vertical", "ascending"),
        ("vertical", "descending"),
        ("radial", "clockwise"),
        ("radial", "counterclockwise"),
    ]:
        make_cascade(
            layout=layout,
            orientation=orientation,
            style={
                "cascade": {
                    "extend": True,
                }
            },
        )


def test_extend_leaflabels():
    for layout, orientation in [
        ("horizontal", "right"),
        ("horizontal", "left"),
        ("vertical", "ascending"),
        ("vertical", "descending"),
        ("radial", "clockwise"),
        ("radial", "counterclockwise"),
    ]:
        make_cascade(
            layout=layout,
            orientation=orientation,
            style={
                "cascade": {
                    "extend": "leaf_labels",
                },
            },
        )



"""
Cogent3 layouts
===============

This example shows how to reproduce `cogent3` layouts using `iplotx`.

.. note::
  `cogent3` uses `plotly` as their default backend, whereas `iplotx` uses `matplotlib`.
  Each backend has pros and cons. In general, `plotly` is more directed towards web
  visualisation thanks to their JavaScript library, whereas `matplotlib` is probably
  more popular and supported outside of web environments.
"""

import cogent3
import matplotlib.pyplot as plt
import iplotx as ipx

tree = cogent3.load_tree("data/tree-with-support.json")

fig, ax = plt.subplots(figsize=(5, 4))
art = ipx.tree(
    tree,
    ax=ax,
    leaf_labels=True,
    style="cogent3",
)


# %%
# Angular layouts (without waypoints, looks more pointy) are supported:

fig, ax = plt.subplots(figsize=(5, 4))
art = ipx.tree(
    tree,
    layout="horizontal",
    ax=ax,
    leaf_labels=True,
    layout_angular=True,
    style="cogent3",
)

# %%
# Vertical and radial layouts are also supported along the same lines:

fig, axs = plt.subplots(3, 1, figsize=(6, 15))
for ax, layout in zip(axs, ("vertical", "radial", "horizontal")):
    art = ipx.tree(
        tree,
        layout=layout,
        layout_orientation="left" if layout == "horizontal" else None,
        ax=ax,
        leaf_labels=True,
        layout_angular=True,
        style="cogent3",
    )
fig.tight_layout()

# %%
# Nonangular trees can be used too:

fig, axs = plt.subplots(3, 1, figsize=(6, 15))
for ax, layout in zip(axs, ("vertical", "radial", "horizontal")):
    art = ipx.tree(
        tree,
        layout=layout,
        layout_orientation="left" if layout == "horizontal" else None,
        ax=ax,
        leaf_labels=True,
        layout_angular=False,
        style="cogent3",
    )
fig.tight_layout()



"""
Support for computing edge paths in 3D.
"""

from typing import (
    Optional,
    Sequence,
)
import numpy as np

from ...typing import (
    Pair,
)


def _compute_edge_segments_straight(
    vcoord_data,
    layout_coordinate_system: str = "cartesian",
    shrink: float = 0,
    **kwargs,
):
    """Compute straight edge path between two vertices, in 3D.

    Parameters:
        vcoord_data: Vertex coordinates in data coordinates, shape (2, 3).
        vpath_fig: Vertex path in figure coordinates.
        vsize_fig: Vertex size in figure coordinates.
        trans: Transformation from data to figure coordinates.
        trans_inv: Inverse transformation from figure to data coordinates.
        layout_coordinate_system: The coordinate system of the layout.
        shrink: Amount to shorten the edge at each end, in figure coordinates.
        **kwargs: Additional keyword arguments (not used).
    Returns:
        A pair with the path and a tuple of angles of exit and entry, in radians.

    """

    if layout_coordinate_system not in ("cartesian"):
        raise ValueError(
            f"Layout coordinate system not supported for straight edges in 3D: {layout_coordinate_system}.",
        )

    segments = [vcoord_data[0], vcoord_data[1]]
    return segments


def _compute_edge_segments(
    *args,
    tension: float = 0,
    waypoints: str | tuple[float, float] | Sequence[tuple[float, float]] | np.ndarray = "none",
    ports: Pair[Optional[str]] = (None, None),
    layout_coordinate_system: str = "cartesian",
    **kwargs,
):
    """Compute the edge path in a few different ways."""
    if (waypoints != "none") and (tension != 0):
        raise ValueError("Waypoints not supported for curved edges.")

    if waypoints != "none":
        raise NotImplementedError("Waypoints not implemented for 3D edges.")
        # return _compute_edge_path_waypoints(
        #    waypoints,
        #    *args,
        #    layout_coordinate_system=layout_coordinate_system,
        #    ports=ports,
        #    **kwargs,
        # )

    if np.isscalar(tension) and (tension == 0):
        return _compute_edge_segments_straight(
            *args,
            layout_coordinate_system=layout_coordinate_system,
            **kwargs,
        )

    raise NotImplementedError("Curved edges not implemented for 3D edges.")
    # return _compute_edge_path_curved(
    #    tension,
    #    *args,
    #    ports=ports,
    #    **kwargs,
    # )



import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import iplotx as ipx


@pytest.fixture
def tree():
    tree = {
        "children": [
            {
                "children": [
                    {},
                    {},
                ],
            },
            {
                "children": [
                    {
                        "children": [
                            {"branch_length": 1.5},
                            {},
                        ],
                    },
                    {
                        "children": [
                            {},
                            {},
                        ],
                    },
                ],
            },
        ],
    }
    tree = ipx.ingest.providers.tree.simple.SimpleTree.from_dict(tree)
    return tree


def test_unsupported_layout(tree):
    with pytest.raises(ValueError):
        ipx.artists.TreeArtist(
            tree,
            layout="unsupported",
        )


def test_get_layout(tree):
    art = ipx.artists.TreeArtist(
        tree,
    )
    assert isinstance(art.get_layout(), pd.DataFrame)


def test_style_subtree(tree):
    art = ipx.artists.TreeArtist(
        tree,
    )
    art.style_subtree(
        [tree.children[0]],
        style={
            "vertex": {"size": 100},
            "edge": {"color": "red"},
        },
    )


def test_style_subtree_sequence(tree):
    art = ipx.artists.TreeArtist(
        tree,
    )
    art.style_subtree(
        [tree.children[0]],
        style=[
            {
                "vertex": {"size": 100},
                "edge": {"color": "red"},
            }
        ],
    )


def test_draw_invisible(tree):
    art = ipx.artists.TreeArtist(
        tree,
    )
    art.set_visible(False)
    fig = plt.figure()
    try:
        assert art.draw(fig.canvas.get_renderer()) is None
    finally:
        plt.close(fig)


def test_cascade_maxdepth(tree):
    for layout, orientation in [
        ("horizontal", "right"),
        ("horizontal", "left"),
        ("vertical", "ascending"),
        ("vertical", "descending"),
        ("radial", "clockwise"),
        ("radial", "counterclockwise"),
    ]:
        with ipx.style.context(
            [
                "tree",
                {
                    "cascade": {"facecolor": {tree.children[0]: "blue"}},
                    "layout": {"orientation": orientation},
                },
            ]
        ):
            ipx.artists.TreeArtist(
                tree,
                layout=layout,
            )


def test_cascade_maxdepth_leaflabels(tree):
    for layout, orientation in [
        ("horizontal", "right"),
        ("horizontal", "left"),
        ("vertical", "ascending"),
        ("vertical", "descending"),
        ("radial", "clockwise"),
        ("radial", "counterclockwise"),
    ]:
        with ipx.style.context(
            [
                "tree",
                {
                    "cascade": {
                        "facecolor": {tree.children[0]: "blue"},
                        "extend": "leaf_labels",
                    },
                    "layout": {"orientation": orientation},
                    "leaf": {"deep": True},
                },
            ]
        ):
            ipx.artists.TreeArtist(
                tree,
                layout=layout,
                leaf_labels=["A", "B", "C", "D", "E", "F"],
            )


def test_cascades_leaflabels_missing(tree):
    with ipx.style.context(
        [
            "tree",
            {
                "cascade": {
                    "facecolor": {tree.children[0]: "blue"},
                    "extend": "leaf_labels",
                },
                "leaf": {"deep": True},
            },
        ]
    ):
        with pytest.raises(ValueError):
            ipx.artists.TreeArtist(
                tree,
            )


def test_edges_angular(tree):
    with ipx.style.context(
        layout_angular=True,
    ):
        ipx.artists.TreeArtist(
            tree,
        )


def test_edges_nowaypoints_bool(tree):
    with ipx.style.context(
        layout_angular=True,
        edge_waypoints=False,
    ):
        ipx.artists.TreeArtist(
            tree,
        )


def test_edges_cmap(tree):
    with ipx.style.context(
        edge_cmap="viridis",
        edge_color=[2, 3, 5],
    ):
        art = ipx.artists.TreeArtist(
            tree,
        )
    # We cannot test the entire thing without plotting
    # because matplotlib delays cmap to rendering time
    # but at least we can check the _A array
    np.testing.assert_array_equal(
        art.get_edges()._A[:6],
        np.ma.MaskedArray(
            [2, 3, 5, 2, 3, 5],
        ),
    )


def test_leafedges_cmap(tree):
    with ipx.style.context(
        leafedge_cmap="viridis",
        leafedge_color=[2, 3, 5],
        leaf_deep=True,
    ):
        art = ipx.artists.TreeArtist(
            tree,
        )
    # We cannot test the entire thing without plotting
    # because matplotlib delays cmap to rendering time
    # but at least we can check the _A array
    np.testing.assert_array_equal(
        art.get_leaf_edges()._A,
        np.ma.MaskedArray(
            [2, 3, 5, 2, 3, 5],
        ),
    )



"""
Branch support
==============

This example shows how to visualise support values for clades or branches.
"""

import cogent3
import matplotlib.pyplot as plt
import iplotx as ipx

tree = cogent3.load_tree("data/tree-with-support.json")

fig, ax = plt.subplots(figsize=(5, 4))
art = ipx.tree(
    tree,
    ax=ax,
    leaf_labels=True,
    style="cogent3",
    show_support=True,
)

# %%
# The location of the support vertex labels can be tuned:

fig, axs = plt.subplots(2, 1, figsize=(5, 9))
art = ipx.tree(
    tree,
    ax=axs[0],
    leaf_labels=True,
    style="cogent3",
    show_support=True,
    vertex_label_hmargin=-1,
    vertex_label_vmargin=5,
    vertex_label_horizontalalignment="right",
    title="Move top left, right-aligned",
)
art = ipx.tree(
    tree,
    ax=axs[1],
    leaf_labels=True,
    style="cogent3",
    show_support=True,
    vertex_label_hmargin=+1,
    vertex_label_vmargin=0,
    vertex_label_horizontalalignment="left",
    title="Move center right, left-aligned",
)

# %%
# .. warning::
#   Using `show_support=True` overrides vertex labels. At the moment it is not
#   possible to set both support values and external vetex labels. Of course,
#   you can craft your own vertex labels to include support values if you want.



"""
Simple directed graph
======================

This example shows how to plot a simple graph with `iplotx`.
"""

import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

g = nx.DiGraph([(0, 1), (1, 2), (2, 3), (3, 4), (4, 0)])
layout = nx.layout.circular_layout(g)
fig, ax = plt.subplots(figsize=(3, 3))
ipx.plot(g, ax=ax, layout=layout)



"""
Knuth Miles US cities
=====================

This example from newtorkx shows an undirected graph of 128 US cities.
"""

import gzip
import re

import matplotlib.pyplot as plt
import networkx as nx
import iplotx as ipx

# Ignore any warnings related to downloading shpfiles with cartopy
import warnings

warnings.simplefilter("ignore")


def miles_graph():
    """Return the cites example graph in miles_dat.txt
    from the Stanford GraphBase.
    """
    # open file miles_dat.txt.gz (or miles_dat.txt)

    fh = gzip.open("data/knuth_miles.txt.gz", "r")

    G = nx.Graph()
    G.position = {}
    G.population = {}

    cities = []
    for line in fh.readlines():
        line = line.decode()
        if line.startswith("*"):  # skip comments
            continue

        numfind = re.compile(r"^\d+")

        if numfind.match(line):  # this line is distances
            dist = line.split()
            for d in dist:
                G.add_edge(city, cities[i], weight=int(d))
                i = i + 1
        else:  # this line is a city, position, population
            i = 1
            (city, coordpop) = line.split("[")
            cities.insert(0, city)
            (coord, pop) = coordpop.split("]")
            (y, x) = coord.split(",")

            G.add_node(city)
            # assign position - Convert string to lat/long
            G.position[city] = (-float(x) / 100, float(y) / 100)
            G.population[city] = float(pop) / 1000
    return G


G = miles_graph()

print("Loaded miles_dat.txt containing 128 cities.")
print(G)

# make new graph of cites, edge if less than 300 miles between them
H = nx.Graph()
for v in G:
    H.add_node(v)
for u, v, d in G.edges(data=True):
    if d["weight"] < 300:
        H.add_edge(u, v)

# nodes colored by degree sized by population
node_color = [float(H.degree(v)) for v in H]

# plot with iplotx
fig, ax = plt.subplots(figsize=(10, 7))
ipx.plot(
    H,
    layout=G.position,
    ax=ax,
    style={
        "vertex": {
            "size": [2 * G.population[v] ** 0.5 for v in H],
            "facecolor": node_color,
            "cmap": "viridis",
        }
    },
)

# %%
# If you compare the style settings with the networkx example, you notice that
# `iplotx` sets the vertex diameter proportional with the "size" parameters,
# while `networkx` follows the `pylab.scatter` convention of scaling the diameter
# of each dot with the **square root** of the chosen size. Therefore, to
# recover the same plot as in the networkx example, we need to scale the
# size manually with the square root of the population.
#
# If we scale the vertex diameter with the population, it would look like this:

fig, ax = plt.subplots(figsize=(10, 7))
ipx.plot(
    H,
    layout=G.position,
    ax=ax,
    style={
        "vertex": {
            "size": [0.1 * G.population[v] for v in H],
            "facecolor": node_color,
            "cmap": "viridis",
        }
    },
)

# %%
# As you can see, the plot is almost identical but larger cities are more prominent.
#
# .. note::
#   There is no right or wrong behaviour here. In fact, not even `matplotlib` is internally consistent
#   with this, and `seaborn` and `matplotlib` use different conventions. If you want to have more dynamic range
#   (i.e. see smaller dots more clearly), scaling with the square root like in `networkx`, or even with the
#   cubic root or logarithm of the size is a good idea. If you want the user to imagine diameters/linear sizes
#   as related to the population, you can use a linear scaling, which is the default in `iplotx`.
#
# If you are curious about why `plotx` scales linearly with the diameter, the explanation is quite simple.
# `networkx` and others take the square root under the assumption that in that way the area of the node marker
# is proportional to the chosen size. However, that works less well for certain shapes, such as crosses, X, or
# lines. In that (not entirely uncommon) case, there is no area as such and the scaling as the square root ends up
# becoming a headache for everyone. This is no criticism of `networkx`, which is great. Just a little backstory.



import pytest
import unittest
import importlib
import matplotlib as mpl

mpl.use("agg")
import matplotlib.pyplot as plt
import iplotx as ipx
from utils import image_comparison

if importlib.util.find_spec("ete4") is None:
    raise unittest.SkipTest("ete4 not found, skipping tests")


@pytest.fixture
def small_tree():
    from ete4 import Tree

    tree = Tree(
        "((),((),(((),()),((),()))));",
    )

    return tree


@image_comparison(baseline_images=["tree_basic"], remove_text=True)
def test_basic(small_tree):
    fig, ax = plt.subplots(figsize=(3, 3))
    ipx.plotting.tree(
        tree=small_tree,
        ax=ax,
        layout="horizontal",
    )


@image_comparison(baseline_images=["tree_radial"], remove_text=True)
def test_radial(small_tree):
    fig, ax = plt.subplots(figsize=(3, 3))
    ipx.plotting.tree(
        tree=small_tree,
        ax=ax,
        layout="radial",
        aspect=1,
    )


@image_comparison(baseline_images=["leaf_labels"], remove_text=True)
def test_leaf_labels(small_tree):
    leaf_labels = {leaf: str(i + 1) for i, leaf in enumerate(small_tree.leaves())}

    fig, ax = plt.subplots(figsize=(4, 4))
    ipx.plotting.tree(
        tree=small_tree,
        ax=ax,
        layout="horizontal",
        vertex_labels=leaf_labels,
        margins=0.1,
    )


@image_comparison(baseline_images=["leaf_labels_hmargin"], remove_text=True)
def test_leaf_labels_hmargin(small_tree):
    leaf_labels = {leaf: str(i + 1) for i, leaf in enumerate(small_tree.leaves())}

    fig, ax = plt.subplots(figsize=(4, 4))
    ipx.plotting.tree(
        tree=small_tree,
        ax=ax,
        layout="horizontal",
        vertex_labels=leaf_labels,
        vertex_label_hmargin=[10, 22],
        margins=(0.15, 0),
    )


@image_comparison(baseline_images=["split_edges"], remove_text=True)
def test_split_edges(small_tree):
    fig, ax = plt.subplots(figsize=(4, 4))
    ipx.plotting.tree(
        tree=small_tree,
        ax=ax,
        layout="horizontal",
        edge_split_linestyle=":",
    )



# These properties are at the bottom of a style dictionary. Values corresponding
# to these keys are rotatable.
rotating_leaves = (
    "cmap",
    "color",
    "size",
    "edgecolor",
    "facecolor",
    "linewidth",
    "linestyle",
    "alpha",
    "zorder",
    "tension",
    "offset",
    "rotate",
    "marker",
    "waypoints",
    "horizontalalignment",
    "verticalalignment",
    "boxstyle",
    "hpadding",
    "vpadding",
    "hmargin",
    "vmargin",
    "ports",
    "width",
    "height",
    "shrink",
    # DEPRECATED
    "padding",
)

# These properties are also terminal style properties, but they cannot be rotated.
# This might change in the future as the API improves.
nonrotating_leaves = (
    "paralleloffset",
    "looptension",
    "loopmaxangle",
    "vertexpadding",
    "extend",
    "deep",
    "angular",
    "curved",
    "capstyle",
)

# Union of all style leaves (rotating and nonrotating)
style_leaves = tuple(list(rotating_leaves) + list(nonrotating_leaves))



"""
Module for handling edge ports in iplotx.
"""

from collections.abc import Callable
import numpy as np

sq2 = np.sqrt(2) / 2

port_dict = {
    "s": (0, -1),
    "w": (-1, 0),
    "n": (0, 1),
    "e": (1, 0),
    "sw": (-sq2, -sq2),
    "nw": (-sq2, sq2),
    "ne": (sq2, sq2),
    "se": (sq2, -sq2),
}


def _get_port_unit_vector(
    portstring: str,
    trans_inv: Callable,
):
    """Get the tangent unit vector from a port string."""
    # The only tricky bit is if the port says e.g. north but the y axis is inverted, in which
    # case the port should go south. We can figure it out by checking the sign of the monotonic
    # trans_inv from figure to data coordinates.
    v12 = trans_inv(
        np.array(
            [
                [0, 0],
                [1, 1],
            ]
        )
    )
    invertx = v12[1, 0] - v12[0, 0] < 0
    inverty = v12[1, 1] - v12[0, 1] < 0

    if invertx:
        portstring = portstring.replace("w", "x").replace("e", "w").replace("x", "e")
    if inverty:
        portstring = portstring.replace("n", "x").replace("s", "n").replace("x", "s")

    if portstring not in port_dict:
        raise KeyError(f"Port not found: {portstring}")
    return np.array(port_dict[portstring])



from typing import (
    Any,
    Optional,
    Sequence,
)
import importlib
from ...typing import (
    TreeDataProvider,
)


class Cogent3DataProvider(TreeDataProvider):
    def preorder(self) -> Sequence[Any]:
        return self.tree.preorder()

    def postorder(self) -> Sequence[Any]:
        return self.tree.postorder()

    def get_leaves(self) -> Sequence[Any]:
        return self.tree.tips()

    @staticmethod
    def get_children(node: Any) -> Sequence[Any]:
        return node.children

    @staticmethod
    def get_branch_length(node: Any) -> Optional[float]:
        return node.length

    @staticmethod
    def check_dependencies() -> bool:
        return importlib.util.find_spec("cogent3") is not None

    @staticmethod
    def tree_type():
        from cogent3.core.tree import PhyloNode

        return PhyloNode

    def get_support(self):
        """Get support values for all nodes."""
        support_dict = {}
        for node in self.preorder():
            support_dict[node] = node.params.get("support", None)
        return support_dict
